from qt import *
import sys

from mdidoc import *
from RoomView import *
from GenerationsGraph import *
from FittnessFrequency import *
from ScrollViewTreeGraph import *

class GenerationControl(QWidget):
    def __init__(self, *args):
        apply(QWidget.__init__, (self,) + args)
        
        document = self.parent().document
        document._generation_control = self
        
        self._generating = False
        self._loading = False
        
        self.settings_button = QPushButton(QString("Settings"), self)
        self.generate_button = QPushButton(QString("Generate!"), self)
        self.slider_label = QLabel(QString("Current generation:"), self)
        self.generation_number_label = QLabel(QString("0"), self)
        self.generations_slider = QSlider(0, document.generation_count, 10, 0, Qt.Horizontal, self)
        
        self.hintSize = QSize(0,self.generate_button.height())
        
        self.slider_label.setAutoResize(True)
        self.generations_slider.setTickmarks(QSlider.Below)
        self.generation_number_label.setAutoResize(True)
        
        self.connect(self.settings_button, SIGNAL("clicked()"), self.clickSettings)
        self.connect(self.generate_button, SIGNAL("clicked()"), self.clickGenerate)
        self.connect(self.generations_slider, SIGNAL("valueChanged(int)"), self.moveSlider)
        
        self.setMinimumSize(self.hintSize)
#        self.setMaximumSize(self.hintSize)
        
        if (document._viewing_generation <> None):
            self._loading = True
            self.generations_slider.setValue(document._viewing_generation)
            self._loading = False
    def sizeHint(self):
        return self.hintSize
    def resizeEvent(self, eventResize):
        space = 10
        height = self.hintSize.height()
        
        width = space + self.settings_button.width()
        self.generate_button.setGeometry(width,0,self.generate_button.width(),height)
        
        width += space + self.generate_button.width()
        self.slider_label.setGeometry(width,height/2-self.slider_label.height()/2,self.slider_label.width(),self.slider_label.height())
        
        width += space + self.slider_label.width()
        self.generation_number_label.setGeometry(width,height/2-self.generation_number_label.height()/2,self.generation_number_label.width(),self.generation_number_label.height())        
        
        width += space +self.generation_number_label.width()
        self.generations_slider.setGeometry(width,0,eventResize.size().width() - width,height)
        
        width += space +self.generations_slider.width()
    def clickSettings(self):
        document = self.parent().document
        settings = document.settings_dialog
        if (settings.exec_loop() <> QDialog.Accepted):
            return
        new_generation_count = document.generation_count
        if (settings.getGenerationCount() > document.generation_count):
            new_generation_count = settings.getGenerationCount()
            self.generate_button.setEnabled(True)
            self.generations_slider.setMaxValue(new_generation_count)
        document.setGPParameters(settings.getCrossoverType(),settings.getPopulationSize(),new_generation_count,settings.getCrossoverPercentage())
    def clickGenerate(self):
        if (self._generating == True):
            self._generating = False
            self.generate_button.setText(QString("Generate!"))
            return
        elif (self._generating == False):
            self._generating = True
            self.generate_button.setText(QString("Cancel"))
#        self.generate_button.setEnabled(False)
        document = self.parent().document
        generator = document.generations
        # FIXME: Set this somewhere
#        step_size = int(document.population_size / 25)
#        if (step_size == 0):
#            step_size = 1
        step_size = 50
        visualisation_control = self.parent().visualisation_control_splitter.visualisation_control
        if (document._generated_generations == None):
            document._generated_generations = 0
        elif (visualisation_control._pause == False):
            visualisation_control.clickPause()
        while (generator.getCurrentGenerationNumber() <= document.generation_count):
            if (generator.getCurrentGenerationNumber() > document._generated_generations):
                document._generated_generations = generator.getCurrentGenerationNumber()
                self.generations_slider.setValue(document._generated_generations)
                generation = generator.getCurrentGenerationNumber()
            fitnesses = generator.runIterativeEvolution(step_size)
            document.addFitnesses(fitnesses)
            document.viewGeneration(document._generated_generations)
            qApp.processEvents()
            if (self._generating == False):
                break
        if (self._generating == False):
            return
        self._generating = False
        self.generate_button.setText(QString("Generate!"))
        self.generate_button.setEnabled(False)
        # INFO: Go to the newest generation
        self.viewGeneration(document._generated_generations)    # INFO: Different to document's viewGeneration!
        visualisation_control.selectIndividual(-1)
        visualisation_control.clickPause()
    def closeEvent(self, e):
        if (self._generating == True):
            self._generating = False
        return QWidget.closeEvent(self, e)
    def moveSlider(self, value):
        document = self.parent().document
        if (document._generated_generations == None):
            return
        if (value > document._generated_generations):
            self.generate_button.setEnabled(True)
            return
        self.generation_number_label.setText(QString(str(value)))
        if (self._generating == True):
            return
        self.viewGeneration(value)
    def viewGeneration(self, n):
        document = self.parent().document
        document.viewGeneration(n)
        if (self._loading == False):
            visualisation_control = self.parent().visualisation_control_splitter.visualisation_control
            visualisation_control.selectGeneration(n)

############################################################
############################################################
############################################################

class VisualisationControl(QWidget):
    def __init__(self, *args):
        apply(QWidget.__init__, (self,) + args)
        
        document = self.parent().parent().document
        document._visualisation_control = self
        
        self.hintSize = QSize(500,250)
        self._pause = True
        self._unpaused_rate = 5
        self._inserting_individuals = False
        
        self.playback_slider = QSlider(-20, 20, 5, 0, Qt.Horizontal, self)
        self.pause_button = QPushButton(QString("Pause"), self)
        self.playback_label = QLabel(QString("Play rate"), self)
        self.FF_label = QLabel(QString("Fast Forward"), self)
        self.RW_label = QLabel(QString("Rewind"), self)
        self.population_list_label = QLabel(QString("Current individual"), self)
        self.population_list = QComboBox(True, self)
        
        self.FF_label.setAutoResize(True)
        self.RW_label.setAutoResize(True)
        self.playback_label.setAutoResize(True)
        self.population_list_label.setAutoResize(True)
        self.playback_slider.setTickmarks(QSlider.Below)
        self.population_list.setAutoCompletion(True)
        self.population_list.lineEdit().setText(QString("Select generation"))
#        self.population_list.setAutoResize(True)
        
        self.connect(self.pause_button, SIGNAL("clicked()"), self.clickPause)
        self.connect(self.playback_slider, SIGNAL("valueChanged(int)"), self.moveSlider)
        self.connect(self.population_list, SIGNAL("activated(int)"), self.selectIndividual)
        self.connect(self.population_list, SIGNAL("textChanged(const QString &)"), self.inputIndividual)
        
#        self.setMinimumSize(self.hintSize)
#        self.setMaximumSize(self.hintSize)
        
        if (document._viewing_individual <> None):
            self.selectGeneration(document._viewing_generation)
            self.population_list.setCurrentItem(document._viewing_individual)
            self.selectIndividual(document._viewing_individual)
    def sizeHint(self):
        return self.hintSize
    def resizeEvent(self, eventResize):
        width = eventResize.size().width()
        height = self.playback_slider.height()
        space = 10
        self.playback_slider.setGeometry(0,0,width,height)
        height += space / 2
        self.pause_button.setGeometry(width / 2 - self.pause_button.width() / 2, height, self.pause_button.width(), self.pause_button.height())
        height += self.pause_button.height() + space
        self.playback_label.setGeometry(width / 2 - self.playback_label.width() / 2, height, self.playback_label.width(), self.playback_label.height())
        self.RW_label.setGeometry(0, height, self.RW_label.width(), self.RW_label.height())
        self.FF_label.setGeometry(width - self.FF_label.width(), height, self.FF_label.width(), self.FF_label.height())
        height += space + self.playback_label.height()
        self.population_list_label.setGeometry(0, height + (self.population_list.height() / 2 - self.population_list_label.height() / 2), self.population_list_label.width(), self.population_list_label.height())
        line_edit_width = 200   # self.population_list.width()
        self.population_list.setGeometry(self.population_list_label.width() + space, height, line_edit_width, self.population_list.height())
    def moveSlider(self, value):
        room_view = self.parent().parent().visualisation_splitter.room_view
        if (value == 0):
            self._pause = True
            self.pause_button.setText(QString("Unpause"))
        else:
            self._unpaused_rate = value
            self._pause = False
            self.pause_button.setText(QString("Pause"))
        rate = float(value) / 100.0
        room_view.setRate(rate)
    def selectIndividual(self, index):
        if (self._inserting_individuals == True):
            return
        individual = None
        document = self.parent().parent().document
        if (index == -1):
            individual = document.generations.getGenerationBest(document._viewing_generation)
            for i in range(len(document._individual_list)):
                item = document._individual_list[i]
                if (item is individual):
                    index = i
                    break
        elif (index < len(document._individual_list)):
            individual = document._individual_list[index]
        else:
            return
        room_view = self.parent().parent().visualisation_splitter.room_view
#        tree = self.parent().parent().visualisation_splitter.visualisation_tools_splitter.tree
        program = individual.getProgram()
        print "Current program:\n", program
#        tree.setNewTree(program)
        initial_robot = None
        if (document._viewing_individual == None):
            initial_robot = Robot()
            print initial_robot
        room_view.setProgram(program, initial_robot)
        current_rate = self.playback_slider.value()
        if (current_rate < 0):
            current_rate = 0
        self.playback_slider.setValue(current_rate)
        self.moveSlider(current_rate)   # INFO: This is required as 'setValue' doesn't call the slot
        if (index < 0):
            document.viewIndividual(None)
        else:
            document.viewIndividual(index)
    def clickPause(self):
        document = self.parent().parent().document
        if (document._viewing_individual == None):  # INFO: Don't do anything if we're not viewing a robot
            return
        if (self._pause == True):
            self.playback_slider.setValue(self._unpaused_rate)
        else:
            self.playback_slider.setValue(0)
    def inputIndividual(self, string):
        if (string.isEmpty() == True):
            return
        result = string.toInt()
        index = result[0]
        result = result[1]
        if ((index >= -1) and (result == True)):
            self.selectIndividual(index)
#        else:
#            self.population_list.lineEdit().setText(QString("Bad index"))
    def selectGeneration(self, value):
        document = self.parent().parent().document
        document._individual_list = []
        for individual in document.generations.getGeneration(value).getIndividuals():
            document._individual_list.append(individual)
        document._individual_list.sort(lambda x, y: y.getFitness() - x.getFitness())    # INFO: Sort fitness descending
        self.population_list.clear()
        self._inserting_individuals = True
        for item in document._individual_list:
            item_string = str(item.getFitness())
            self.population_list.insertItem(QString(item_string))
        self._inserting_individuals = False

###########################

class VisualisationToolsSplitter(QSplitter):
    def __init__(self, *args):
        apply(QSplitter.__init__, (self,) + args)
        self.setOrientation(Qt.Vertical)
        self.children = []
    def addChildren(self):
        document = self.parent().parent().document
        self.histogram = FittnessFrequency(document, self)
        self.children.append(self.histogram)
        self.tree = ScrollViewTreeGraph(document, self)
        self.children.append(self.tree)
        
        self.parent().room_view.setUpdatePartner(document)
        
        self.setResizeMode(self.histogram, QSplitter.KeepSize)
    def closeEvent(self, eventClose):
        for child in self.children:
            child.close()
        QWidget.closeEvent(self, eventClose)

class VisualisationSplitter(QSplitter):
    def addChildren(self):
        self.children = []
        document = self.parent().document
        self.room_view = RoomView(document, self)
        self.children.append(self.room_view)
        self.visualisation_tools_splitter = VisualisationToolsSplitter(self)
        self.visualisation_tools_splitter.addChildren()
        self.children.append(self.visualisation_tools_splitter)
    def closeEvent(self, eventClose):
        for child in self.children:
            child.close()
        QWidget.closeEvent(self, eventClose)

class VisualisationControlSplitter(QSplitter):
    def addChildren(self):
        self.children = []
        self.visualisation_control = VisualisationControl(self)
        self.children.append(self.visualisation_control)
        document = self.parent().document
        self.generations_graph = GenerationsGraph(document, self)
        self.children.append(self.generations_graph)
        
        document._main_generation_graph = self.generations_graph
        
        self.setResizeMode(self.visualisation_control, QSplitter.KeepSize)
    def closeEvent(self, eventClose):
        for child in self.children:
            child.close()
        QWidget.closeEvent(self, eventClose)

class ChildFrame(QSplitter):
    def __init__(self, document, *args):
        self.document = document
        apply(QSplitter.__init__, (self,) + args)
        self.setOrientation(Qt.Vertical)
        self.children = []
        self.addChildren()
    def addChildren(self):
        self.generation_control = GenerationControl(self)
        self.children.append(self.generation_control)
        self.visualisation_splitter = VisualisationSplitter(self)
        self.visualisation_splitter.addChildren()
        self.children.append(self.visualisation_splitter)
        self.visualisation_control_splitter = VisualisationControlSplitter(self)
        self.visualisation_control_splitter.addChildren()
        self.children.append(self.visualisation_control_splitter)
        
        self.setResizeMode(self.generation_control, QSplitter.KeepSize)
        self.setResizeMode(self.visualisation_control_splitter, QSplitter.KeepSize)
    def closeEvent(self, eventClose):
        for child in self.children:
            child.close()
        QWidget.closeEvent(self, eventClose)

if __name__ == "__main__":
    app = QApplication(sys.argv)
    frame = ChildFrame()
    app.setMainWidget(frame)
    frame.addChildren()
    frame.show()
    app.exec_loop()
