"""
FIXMEs:
-   Topmost (last) Y-axis label overlapping with previous
"""
from math import *
from qt import *

from GP_Imp import *

class FittnessFrequency(QWidget):
    def __init__(self, document=None, parent=None, *args):
        self._document = document
        apply(QWidget.__init__,(self, parent) + args)
        self.background_color = QColor(255,255,255)
        self.setEraseColor(self.background_color)

        self.height = 0
        self.width = 0
        self.num_bars = 16
        
        self.font_family =  'Courier' 
#        self.font_size = 6
        self.font_height = 9
        self.font_weight  = QFont.Normal
        self.font_italic = False
        
#        self.scale_font = QFont(self.font_family,self.font_size, self.font_weight, self.font_italic)
        self.scale_font = QFont(self.font_family)
        self.scale_font.setPixelSize(self.font_height)
        
        self.axis_space = 30
        self.label_space = 20
        self.margin = 50
        self.extra_space = 10
        self.extra_font_height = 10
        
        self.title = "Fittness Frequency"
        self.x_axis_label = "Fittness"
        self.y_axis_label = "Frequency"
        
        self.x_max_value = 320
        self.y_max_value = 1    # INFO: Default height (pretty sensible value, can't have 0)
        
        self.xDiff = 0.0
        self.yDiff = 0.0
        
        self.pixmap = QPixmap(1,1)
        self.painter = QPainter(self.pixmap)
        
        self.hintSize = QSize(300, 300)
        
        self.resetData()
        
        self.connect(self._document, PYSIGNAL("sigViewGeneration"), self.showGeneration)
        
        if (self._document._viewing_generation <> None):
            self.showGeneration(self._document._viewing_generation)
    
    def showGeneration(self, generation):
        fitnesses = self._document.generations.getFitnessesOfGeneration(generation)
        self.resetData()
        self.addMoreData(fitnesses)
    
    def resetData(self):
        self.frequency_dict = {}
        for i in range(self.x_max_value + 1):
            self.frequency_dict[i] = 0
#        self.pixmapDraw()
    
    def sizeHint(self):
        return self.hintSize
    
    def paintHistogram(self):
        # FIXME: Check whether view is closing
        if (self.painter.isActive() == False):
            self.painter.begin(self)
        self.drawHistogram()
        self.painter.end()
    
    def pixmapDraw(self):
        self.painter = QPainter(self.pixmap)
        self.pixmap.fill(self.background_color)
        self.paintHistogram()
        self.paintMe()
    
    def addMoreData(self, new_freq_list):
        self.addData(new_freq_list)
        self.pixmapDraw()
    
    def setBarCount(self, num_bars):
        self.num_bars = num_bars
        self.pixmapDraw()
    
    def paintEvent(self, ev):
        self.paintMe()
    
    def paintMe(self):
        bitBlt(self, 0, 0, self.pixmap)
    
    def resizeEvent(self, eventResize):
        new_size = eventResize.size()
        self.width = new_size.width()
        self.height = new_size.height()
        if ((self.width == 0) or (self.height == 0)):
            return
        self.pixmap = QPixmap(self.width,self.height)
        self.pixmapDraw()
    
    def drawHistogram(self):
        self.setFrequencyValues()
        self.y_max_value = max(self.bar_fittness_freq_list)
        if (self.y_max_value == 0): # INFO: If there's no data
            self.y_max_value = 1
        self.yDiff = float(self.height - 2*self.margin) / float(self.y_max_value)
        if (self.yDiff <= 0.0):
            return
        self.drawHistogramAxes()
        self.drawLabels()
        self.drawXScales()
        self.drawYScales()
        self.drawBars()
    
    def drawHistogramAxes(self):
        self.drawAxis(self.margin, self.height-self.margin, self.width-self.margin + self.extra_space, self.height-self.margin)
        self.drawAxis(self.margin, self.height-self.margin, self.margin, self.margin - self.extra_space)
    
    def drawAxis(self, xFrom, yFrom, xTo, yTo):
        self.painter.drawLine(xFrom,yFrom,xTo,yTo)
    
    def drawLabels(self):
        self.painter.drawText(0, 0, self.width, self.label_space, Qt.AlignCenter, self.title)
        self.painter.drawText(0, self.height-self.label_space, self.width, self.label_space, Qt.AlignCenter, self.x_axis_label)
        self.painter.drawText(self.label_space, self.label_space, self.width, self.label_space, Qt.AlignVCenter, self.y_axis_label)
    
    def drawXScales(self):
        x = self.margin
        y = self.height - self.axis_space
        self.xDiff = (self.width - 2*self.margin)/self.num_bars
        for value in range(0, self.x_max_value+1, self.x_max_value/self.num_bars):
            self.drawScale(x,y,value)
            x += self.xDiff
    
    def drawYScales(self):
        x = self.axis_space
        y = self.height - self.margin
        font_height = self.scale_font.pixelSize() + self.extra_font_height
        step = float(font_height)
        accum = 0.0
        for value in range(0, self.y_max_value+1):
            if (accum == 0.0):
                self.drawScale(x,y,value)
            elif (value == self.y_max_value):
                y = self.height - self.margin - ((float(value) / float(self.y_max_value+1)) * float(self.height - 2*self.margin))
                self.drawScale(x,y,value)
            accum += self.yDiff
            if (accum >= step):
                y = self.height - self.margin - ((float(value) / float(self.y_max_value+1)) * float(self.height - 2*self.margin))
                accum = 0.0
    
    def drawScale(self, x, y, value):
        self.painter.setFont(self.scale_font)
        self.painter.drawText(x, y, str(value))
    
    def drawBars(self):
        max_bar_fittness = self.x_max_value/self.num_bars
        x = self.margin
        y = self.height - self.margin       
        bar_num = 1
        for freq in self.bar_fittness_freq_list:
            y = self.height - self.margin - int(float(freq)*self.yDiff)
            self.drawBar(x, y, self.xDiff, int(float(freq)*self.yDiff), bar_num)
            x += self.xDiff
            y = self.height - self.margin
            bar_num += 1
    
    def drawBar(self, x, y, w, h, bar_num):
        r = 0
        if self.yDiff > 0.0:
            r = float(h/(float(self.y_max_value)*self.yDiff))*255
        g = bar_num * float(255/self.num_bars)
        b = (self.num_bars-bar_num+1) * float(255/self.num_bars)
        self.painter.setBrush(QBrush(QColor(r,g,b)))
        self.painter.drawRect(x,y,w,h)
    
    def addData(self, list):
        for item in list:
            if item <= self.x_max_value:
                self.frequency_dict[item] += 1
    
    def setFrequencyValues(self):
        max_bar_fittness = (self.x_max_value + 1)/self.num_bars
        bar_fittness = max_bar_fittness
        freq_count = 0
        self.bar_fittness_freq_list = []
        for key in range(self.x_max_value+1):
            if key < bar_fittness:
                freq_count += self.frequency_dict[key]
            elif key == bar_fittness:
                self.bar_fittness_freq_list.append(freq_count)
                freq_count = self.frequency_dict[key]
                bar_fittness += max_bar_fittness
        new_length = len(self.bar_fittness_freq_list)
        if ((new_length > self.num_bars) and (new_length > 1)):
            extra = self.bar_fittness_freq_list[-1]
            self.bar_fittness_freq_list = self.bar_fittness_freq_list[:-1]
            self.bar_fittness_freq_list[-1] += extra

if __name__ == "__main__":
    import sys
    app=QApplication(sys.argv)
    fittnessFrequency = FittnessFrequency()
    fittnessFrequency.addMoreData([1,2,3,4,33,33,3,3,33,80, 100,100,100,100,100,100,100,100,161,90,91,91,160, 160 ,198,199,200,231,240, 240, 240, 257, 300,301,302,320, 320, 320, 320])
    fittnessFrequency.addMoreData([1,2,3,100,120,102,120,120,120,120,120,120])
    list = [100] * 100
    fittnessFrequency.addMoreData(list)
#    fittnessFrequency.setBarCount(16)
    app.setMainWidget(fittnessFrequency)
    fittnessFrequency.show()
    app.exec_loop()
