# COMP3141 Assignment 2
#
# Wall-Following Robot GP Implementation
#
# Balint Seeber & Janice Leung

from random import Random
#import psyco
#psyco.full()

random = Random()

TRUE    = 1
FALSE   = 0
EMPTY   = FALSE
WALL    = TRUE

SINGLE_CROSSOVER = 1
DOUBLE_CROSSOVER = 2

ValidPositions = (
    (1,1),(1,2),(1,3),(1,4),(1,5),(1,6),(1,7),(1,8),
    (2,1),(2,2),(2,3),(2,4),(2,5),(2,6),(2,7),(2,8),
    (3,1),(3,2),(3,3),(3,4),(3,5),(3,6),(3,7),(3,8),
    (4,1),(4,2),(4,3),(4,4),(4,5),(4,6),
    (5,1),(5,2),(5,3),(5,4),(5,5),(5,6),
    (6,1),(6,2),(6,3),(6,4),(6,5),(6,6),(6,7),(6,8),
    (7,1),(7,2),(7,3),(7,4),(7,5),(7,6),(7,7),(7,8),
    (8,1),(8,2),(8,3),(8,4),(8,5),(8,6),(8,7),(8,8))

ValidPositionCount = len(ValidPositions)

class ProgramNode:
    def __eq__(self, other):
        return str(self) == str(other)
    def countNodes(self):
        self.node_count = 1
        if self.hasChildren():
            for node in self.nodes:
                self.node_count += node.countNodes()
        return self.node_count
    def pickSubtree(self, crossover_point):
        if(crossover_point == 1):
            return self
        count = 1
        for node in self.nodes:
            node_count = node.countNodes()
            if crossover_point <= (count + node_count):
                return node.pickSubtree(crossover_point-count)
            count += node_count
    def replaceSubtree(self, crossover_point, new_program):
        (program,flag) = self.replaceSubtreeFast(crossover_point, new_program)
        return program
    def replaceSubtreeFast(self, crossover_point, new_program):
        if (crossover_point == 1):
            return (new_program,TRUE)
        return (self.__class__(),FALSE)
    def singleCrossover(self, self_crossover, other_program, other_crossover):
        sub_tree = other_program.pickSubtree(other_crossover)
        new_tree = self.replaceSubtree(self_crossover, sub_tree)
        return new_tree
    def doubleCrossover(self, self_crossover, other_program, other_crossover):
        sub_tree1 = self.pickSubtree(self_crossover)
        sub_tree2 = other_program.pickSubtree(other_crossover)
        new_tree1 = self.replaceSubtree(self_crossover, sub_tree2)
        new_tree2 = other_program.replaceSubtree(other_crossover, sub_tree1)
        return (new_tree1,new_tree2)
    def getChildren(self):
        self.children_list = []
        if self.hasChildren():
            for node in self.nodes:
                self.children_list.append(node)
        return self.children_list
    def copy(self):
        return self.__class__()
    def newRandomNode(self):
        pick = random.randint(0, NodeListLength - 1)
        return NodeList[pick].copy()
    def randomiseNodes(self):
        return
    def hasChildren(self):
        if isinstance(self, And) or isinstance(self, Or) or isinstance(self, Not) or isinstance(self, If):
            return TRUE
    ###############
    ### our functions ###
    ###############
    def resetData(self):
        self.x = 0
        self.y = 0
        self.level = -1
        self.parent = None
        self.max_height = 1
        for child in self.getChildren():
            child.resetData()
    def getTreeHeight(self):
        if self.getChildren() == []:
            return 1
        children_height_list = []
        for child in self.getChildren():
            children_height_list.append(child.getTreeHeight())
        max_height_in_list = max(children_height_list)
        self.max_height += max_height_in_list
        return self.max_height
    def getX(self):
        return self.x
    def getY(self):
        return self.y
    def setX(self, x):
        self.x = x
    def setY(self, y):
        self.y = y
    def getLevel(self):
        return self.level
    def setLevel(self, level):
        self.level = level

class BooleanValue(ProgramNode):
    def __str__(self):
        return self.__class__.__name__ + "()"

class One(BooleanValue):
    def evalAtPosition(self, position):
        return TRUE

class Zero(BooleanValue):
    def evalAtPosition(self, position):
        return FALSE


class And(ProgramNode):
    def __init__(self, p1, p2):
        self.nodes = (p1,p2)
    def __str__(self):
        return "And("+str(self.nodes[0])+","+str(self.nodes[1])+")"
    def randomiseNodes(self):
        self.nodes = (self.newRandomNode(), self.newRandomNode())
        for p in self.nodes:
            p.randomiseNodes()
    def evalAtPosition(self, position):
        result1 = self.nodes[0].evalAtPosition(position)
        if (result1 <> FALSE) and (result1 <> TRUE):
            return result1
        elif (result1 == FALSE):
            return FALSE
        return self.nodes[1].evalAtPosition(position)
    def copy(self):
        return And(self.nodes[0].copy(),self.nodes[1].copy())
    def replaceSubtreeFast(self, crossover_point, new_program):
        if (crossover_point == 1):
            return (new_program,TRUE)
        node_count = self.nodes[0].countNodes()
        arg1 = None
        flag1 = FALSE
        if crossover_point > (node_count + 1):
            arg1 = self.nodes[0].copy()
        else:
            (arg1,flag1) = self.nodes[0].replaceSubtreeFast(crossover_point-1, new_program)
        arg2 = None
        return_flag = FALSE
        if flag1 == FALSE:
            (arg2,flag2) = self.nodes[1].replaceSubtreeFast(crossover_point-node_count-1, new_program)
            return_flag = flag2
        else:
            return_flag = TRUE
            arg2 = self.nodes[1].copy()
        return (And(arg1,arg2),return_flag)

class Or(ProgramNode):
    def __init__(self, p1, p2):
        self.nodes = (p1,p2)
    def __str__(self):
        return "Or("+str(self.nodes[0])+","+str(self.nodes[1])+")"
    def randomiseNodes(self):
        self.nodes = (self.newRandomNode(), self.newRandomNode())
        for p in self.nodes:
            p.randomiseNodes()
    def evalAtPosition(self, position):
        result1 = self.nodes[0].evalAtPosition(position)
        if result1 == TRUE:
            return TRUE
        if (result1 <> FALSE):
            return result1
        return self.nodes[1].evalAtPosition(position)
    def copy(self):
        return Or(self.nodes[0].copy(),self.nodes[1].copy())
    def replaceSubtreeFast(self, crossover_point, new_program):
        if (crossover_point == 1):
            return (new_program,TRUE)
        node_count = self.nodes[0].countNodes()
        arg1 = None
        flag1 = FALSE
        if crossover_point > (node_count + 1):
            arg1 = self.nodes[0].copy()
        else:
            (arg1,flag1) = self.nodes[0].replaceSubtreeFast(crossover_point-1, new_program)
        arg2 = None
        return_flag = FALSE
        if flag1 == FALSE:
            (arg2,flag2) = self.nodes[1].replaceSubtreeFast(crossover_point-node_count-1, new_program)
            return_flag = flag2
        else:
            return_flag = TRUE
            arg2 = self.nodes[1].copy()
        return (Or(arg1,arg2),return_flag)

class Not(ProgramNode):
    def __init__(self, p):
        self.nodes = (p,)
    def __str__(self):
        return "Not("+str(self.nodes[0])+")"
    def randomiseNodes(self):
        self.nodes = (self.newRandomNode(),)
        for p in self.nodes:
            p.randomiseNodes()
    def evalAtPosition(self, position):
        result = self.nodes[0].evalAtPosition(position)
        if (result <> FALSE) and (result <> TRUE):
            return result
        return (TRUE - result)
    def copy(self):
        return Not(self.nodes[0].copy())
    def replaceSubtreeFast(self, crossover_point, new_program):
        if (crossover_point == 1):
            return (new_program,TRUE)
        (arg,flag) = self.nodes[0].replaceSubtreeFast(crossover_point-1, new_program)
        return (Not(arg),flag)

class If(ProgramNode):
    def __init__(self, p1, p2, p3):
        self.nodes = (p1,p2,p3)
    def __str__(self):
        return "If("+str(self.nodes[0])+","+str(self.nodes[1])+","+str(self.nodes[2])+")"
    def randomiseNodes(self):
        self.nodes = (self.newRandomNode(), self.newRandomNode(), self.newRandomNode())
        for p in self.nodes:
            p.randomiseNodes()
    def evalAtPosition(self, position):
        result1 = self.nodes[0].evalAtPosition(position)
        if (result1 <> FALSE) and (result1 <> TRUE):
            return result1
        if result1 == TRUE:
            return self.nodes[1].evalAtPosition(position)
        else:
            return self.nodes[2].evalAtPosition(position)
    def copy(self):
        return If(self.nodes[0].copy(),self.nodes[1].copy(),self.nodes[2].copy())
    def replaceSubtreeFast(self, crossover_point, new_program):
        if (crossover_point == 1):
            return (new_program,TRUE)
        node_count1 = self.nodes[0].countNodes()
        arg1 = None
        flag1 = FALSE
        if crossover_point > (node_count1 + 1):
            arg1 = self.nodes[0].copy()
        else:
            (arg1,flag1) = self.nodes[0].replaceSubtreeFast(crossover_point-1, new_program)
        arg2 = None
        arg3 = None
        return_flag = FALSE
        if flag1 == FALSE:
            flag2 = FALSE
            node_count2 = self.nodes[1].countNodes()
            if crossover_point > (node_count1 + node_count2 + 1):
                arg2 = self.nodes[1].copy()
                (arg3,flag3) = self.nodes[2].replaceSubtreeFast(crossover_point-node_count1-node_count2-1, new_program)
                return_flag = flag3
            else:
                (arg2,flag2) = self.nodes[1].replaceSubtreeFast(crossover_point-node_count1-1, new_program)
                if flag2 == FALSE:
                    (arg3,flag3) = self.nodes[2].replaceSubtreeFast(crossover_point-node_count1-node_count2-1, new_program)
                    return_flag = flag3
                else:
                    arg3 = self.nodes[2].copy()
                    return_value = TRUE
            return_flag = flag2
        else:
            return_flag = TRUE
            arg2 = self.nodes[1].copy()
            arg3 = self.nodes[2].copy()
        return (If(arg1,arg2,arg3),return_flag)


class ActionNode(ProgramNode):
    def __str__(self):
        return self.__class__.__name__ + "()"

class North(ActionNode):
    def evalAtPosition(self, position):
        return (0,-1)

class East(ActionNode):
    def evalAtPosition(self, position):
        return (1,0)

class South(ActionNode):
    def evalAtPosition(self, position):
        return (0,1)

class West(ActionNode):
    def evalAtPosition(self, position):
        return (-1,0)


def IsMovementPossible(position, movement):
        x = position[0] + movement[0]
        y = position[1] + movement[1]
        if ((x == 0) or (y == 0) or (x == 9) or (y == 9)) or (((x == 4) or (x == 5)) and (y > 6)):
            return WALL
        return EMPTY

class SensorNode(ProgramNode):
    def __str__(self):
        return self.__class__.__name__ + "()"

class N(SensorNode):
    def evalAtPosition(self, position):
        return IsMovementPossible(position,(0,-1))

class NE(SensorNode):
    def evalAtPosition(self, position):
        return IsMovementPossible(position,(1,-1))

class E(SensorNode):
    def evalAtPosition(self, position):
        return IsMovementPossible(position,(1,0))

class SE(SensorNode):
    def evalAtPosition(self, position):
        return IsMovementPossible(position,(1,1))

class S(SensorNode):
    def evalAtPosition(self, position):
        return IsMovementPossible(position,(0,1))

class SW(SensorNode):
    def evalAtPosition(self, position):
        return IsMovementPossible(position,(-1,1))

class W(SensorNode):
    def evalAtPosition(self, position):
        return IsMovementPossible(position,(-1,0))

class NW(SensorNode):
    def evalAtPosition(self, position):
        return IsMovementPossible(position,(-1,-1))

LonelyZero = Zero()

NodeList = [And(LonelyZero,LonelyZero), Or(LonelyZero,LonelyZero), Not(LonelyZero), If(LonelyZero,LonelyZero,LonelyZero), One(), Zero(), North(), East(), South(), West(), N(), NE(), E(), SE(), S(), SW(), W(), NW()]

NodeListLength = len(NodeList)

class Robot:
    def __init__(self, x=None, y=None):
        if (x == None or y == None):
            valid_position = ValidPositions[random.randint(0,ValidPositionCount - 1)]
            self._x = valid_position[0]
            self._y = valid_position[1]
        else:
            self._x = x
            self._y = y
    def __eq__(self, other):
        if (other == None):
            return FALSE
        return self._x == other.getX() and self._y == other.getY()
    def __str__(self):
        return "Robot("+str(self._x)+","+str(self._y)+")"
    def getX(self):
        return self._x
    def getY(self):
        return self._y
    def move(self, program):
        movement = program.evalAtPosition((self._x,self._y))
        if (movement <> FALSE) and (movement <> TRUE):
            if IsMovementPossible((self._x,self._y), movement) == EMPTY:
                self._x = self._x + movement[0]
                self._y = self._y + movement[1]
    def getSensorValue(self, sensor_node):
        return sensor_node.evalAtPosition((self.getX(),self.getY()))


class RobotList:
    def __init__(self):
        self._list = []
    def addRobot(self, robot):
        self._list.append(robot)
    def numberUniquePositionsNextToWall(self):
        count = 0
        accum_list = []
        for robot in self._list:
            if(not (robot in accum_list)):
                accum_list.append(robot)
                x = robot.getX()
                y = robot.getY()
                if (x == 1 and y == 1) or (x == 1 and y == 8)or (x == 3 and y == 8)or (x == 6 and y == 8)or (x == 8 and y == 8)or (x == 8 and y == 1):
                    count += 1
                elif (y == 1)or (x == 1) or (x == 8)or (y == 8 and (x <> 4 or x <> 5))or(y==6 and (x == 4 or x == 5))or(y==7 and(x==4 or x==5)) or (x==3 and y==6) or (x==6 and y ==6):
                    count += 1
        return count
    def count(self):
        return len(self._list)
    def getRobots(self):
        return self._list

class Individual:
    def __init__(self, program=LonelyZero, initial_robot=None, fitness=-1):
        self._program = program
        self._fitness = fitness
        self._robot_list = RobotList()
        if not initial_robot:
            initial_robot = Robot()
        self._robot_list.addRobot(initial_robot)
    def __str__(self):
        return "Individual("+str(self._program)+","+str(self._fitness)+")"
    def getProgram(self):
        return self._program
    def getNthRobot(self, n):
        if (n < self._robot_list.count()):
            return self._robot_list.getRobots()[n]
        prev_robot = self._robot_list.getRobots()[self._robot_list.count() - 1]
        run_robot = Robot(prev_robot.getX(), prev_robot.getY())
        runs = (n + 1) - self._robot_list.count()
        for i in range(runs):
            run_robot.move(self._program)
            new_robot = Robot(run_robot.getX(),run_robot.getY())
            if new_robot == prev_robot:
                return new_robot
            prev_robot = new_robot
            self._robot_list.addRobot(new_robot)
        return self._robot_list.getRobots()[n]
    def getFitness(self):
        if self._fitness == -1:
            self._fitness = 0
            for i in range(10):
                ind = Individual(self._program)
                ind.getNthRobot(60)
                self._fitness += ind._robot_list.numberUniquePositionsNextToWall()
        return self._fitness
    def generateRandomProgram(self):
        pick = random.randint(0,3)
        self._program = NodeList[pick].copy()
        self._program.randomiseNodes()

class Population:
    def __init__(self, number=0):
        self._individuals = []
    def newRandomIndividuals(self, number=1):
        list = []
        for i in range(number):
            new_ind = Individual()
            new_ind.generateRandomProgram()
            list.append(new_ind)
            self.addIndividual(new_ind)
        return list
    def mostFitIndividual(self):
        most_fit_index = -1
        most_fit = -1
        for i in range(len(self._individuals)):
            current_fitness = self._individuals[i].getFitness()
            if current_fitness > most_fit:
                most_fit_index = i
                most_fit = current_fitness
        return self._individuals[most_fit_index]
    def pickIndividualsUsingTournamentSelection(self, number=1):
        new_pop = Population()
        for i in range(number):
            sub_pop = Population()
            for j in range(7):
                pick = random.randint(0,len(self._individuals)-1)
                ind = self._individuals[pick]
                sub_pop.addIndividual(ind)
            new_pop.addIndividual(sub_pop.mostFitIndividual())
        return new_pop
    def reproduceIndividuals(self, number=1):
        return self.pickIndividualsUsingTournamentSelection(number)
    def crossoverIndividualsSingle(self, number=1):
        pop = Population()
        father_pop = self.pickIndividualsUsingTournamentSelection(number)
        mother_pop = self.pickIndividualsUsingTournamentSelection(number)
        for i in range(number):
            father = father_pop._individuals[i]
            mother = mother_pop._individuals[i]
            father_program = father.getProgram()
            mother_program = mother.getProgram()
            father_pick = random.randint(1, father_program.countNodes())
            mother_pick = random.randint(1, mother_program.countNodes())
            child = mother_program.singleCrossover(mother_pick, father_program, father_pick)
            ind = Individual(child)
            pop.addIndividual(ind)
        return pop
    def crossoverIndividualsDouble(self, number=1):
        odd_flag = 0
        if (number % 2) == 1:
            odd_flag = 1
            number += 1
        father_pop = self.pickIndividualsUsingTournamentSelection(number/2)
        mother_pop = self.pickIndividualsUsingTournamentSelection(number/2)
        pop = Population()
        for i in range(number/2):
            father = father_pop._individuals[i]
            mother = mother_pop._individuals[i]
            father_program = father.getProgram()
            mother_program = mother.getProgram()
            father_pick = random.randint(1, father_program.countNodes())
            mother_pick = random.randint(1, mother_program.countNodes())
            (child1, child2) = mother_program.doubleCrossover(mother_pick, father_program, father_pick)
            ind1 = Individual(child1)
            pop.addIndividual(ind1)
            if (i <> ((number / 2) - 1)) or (odd_flag == 0):
                ind2 = Individual(child2)
                pop.addIndividual(ind2)
        return pop
    def crossoverIndividuals(self, number=1, type=DOUBLE_CROSSOVER):
        if (type == SINGLE_CROSSOVER):
            return self.crossoverIndividualsSingle(number)
        elif (type == DOUBLE_CROSSOVER):
            return self.crossoverIndividualsDouble(number)
        return Population()
    def addIndividual(self, individual):
        self._individuals.append(individual)
    def addIndividuals(self, population):
        self._individuals.extend(population._individuals)
    def getIndividuals(self):
        return self._individuals

class Generation:
    def __init__(self, initial_population):
        self._generations = [initial_population]
    def getGeneration(self, n):
        if n < len(self._generations):
            return self._generations[n]
        runs = (n + 1) - len(self._generations)
        for i in range(runs):
            new_pop = Population()
            prev_pop = self._generations[len(self._generations) - 1]
            head_count = len(prev_pop._individuals)
            reproduce_number = head_count / 10
            list1 = prev_pop.reproduceIndividuals(reproduce_number)
            new_pop.addIndividuals(list1)
            list2 = prev_pop.crossoverIndividuals(head_count - reproduce_number)
            new_pop.addIndividuals(list2)
            self._generations.append(new_pop)
        return self._generations[n]
    def getGenerationBest(self, n):
        return self.getGeneration(n).mostFitIndividual()


def run_generations(generation_limit):
    print "Creating initial population...\n"
    initial_population = Population()
    initial_population.newRandomIndividuals(5000)
    generation = Generation(initial_population)
    fittest = None
    for i in range(generation_limit):
        print "Starting generation", i + 1
        print "----------------------\n"
        generation.getGeneration(i+1)
    print "Getting fittest:\n"
    fittest = generation.getGenerationBest(i+1)
    print fittest
    print "\n==== FINISHED ====\n"
    return fittest

if __name__ == "__main__":
    pop = Population()
    for i in range(5):
        print "==== PERFORMING RUN", i + 1, "====\n"
        pop.addIndividual(run_generations(10))
    print "GRAND FINALE:\n"
    print pop.mostFitIndividual()

