from GP_Imp import *

class IterativeGeneration:
    def __init__(self, population_size, reproduction_count, crossover_count, crossover_type=DOUBLE_CROSSOVER):
        self._current_population_size = 0
        self._iterative_initial_population = Population()
        self._current_generation = 0
        self._current_individual = 0
        self.setParameters(population_size, reproduction_count, crossover_count, crossover_type)
        self._generations = []
        self._current_population = self._iterative_initial_population
        self._previous_population = None
    
    def setParameters(self, population_size, reproduction_count, crossover_count, crossover_type):
        self._population_size = population_size
        self._reproduction_count = reproduction_count
        self._crossover_count = crossover_count
        self._crossover_type = crossover_type
    
    def getCurrentGenerationNumber(self):
        return self._current_generation

    def getCurrentPopulation(self):
        return self._current_population

    def getGeneration(self, n):
        if (n == 0):
            return self._iterative_initial_population
        try:
            return self._generations[n - 1]
        except:
            return Population()
    
    def getFitnessesOfGeneration(self, n):
        population = self.getGeneration(n)
        list = []
        for individual in population.getIndividuals():
            list.append(individual.getFitness())
        return list

    def getGenerationBest(self, n):
        return self.getGeneration(n).mostFitIndividual()

    def createInitialPopulation(self, step_size=1):
        if ((self._current_population_size + step_size) > self._population_size):
            step_size = self._population_size - self._current_population_size
        new_programs = self._iterative_initial_population.newRandomIndividuals(step_size)
        fitnesses = []
        for program in new_programs:
            fitnesses.append(program.getFitness())
        self._current_population_size += len(fitnesses)
        return fitnesses

    def runIterativeEvolution(self, step_size):
        if (self._current_generation == 0):
            fitnesses = self.createInitialPopulation(step_size)
            if (self._current_population_size == self._population_size):
                print "Initial population complete (", len(self._iterative_initial_population.getIndividuals()), "individuals )"
                self._current_generation = 1
            return fitnesses
        else:
            if (self._current_individual == 0):
                self._previous_population = self._current_population
                self._current_population = Population()
                self._generations.append(self._current_population)
            if ((self._current_individual + step_size) > self._population_size):
                step_size = self._population_size - self._current_individual
            fitnesses = []
            reproduction_step = 0
            crossover_step = 0
            if (self._current_individual < self._reproduction_count):
                if ((self._current_individual + step_size) > self._reproduction_count):
                    reproduction_step = self._reproduction_count - self._current_individual
                    crossover_step = step_size - reproduction_step
                else:
                    reproduction_step = step_size
            else:
                crossover_step = step_size
            if (reproduction_step > 0):
                new_population = self._previous_population.reproduceIndividuals(reproduction_step)
                for individual in new_population.getIndividuals():
                    fitnesses.append(individual.getFitness())
                self._current_population.addIndividuals(new_population)
            if (crossover_step > 0):
                new_population = self._previous_population.crossoverIndividuals(crossover_step, self._crossover_type)
                i = 0
                for individual in new_population.getIndividuals():
                    i += 1
                    if (i > crossover_step):
                        break
                    self._current_population.addIndividual(individual)
                    fitnesses.append(individual.getFitness())
            self._current_individual += len(fitnesses)
            if (self._current_individual == self._population_size):
                print "Finished generation", self._current_generation, "(", len(self.getGeneration(self._current_generation).getIndividuals()), "individuals )"
                self._current_individual = 0
                self._current_generation += 1
            return fitnesses

if __name__ == "__main__":
    population_count = 5000
    reproduction_count = population_count / 10
    crossover_count = population_count - reproduction_count
    generator = IterativeGeneration(population_count, reproduction_count, crossover_count)
    last_generation = 0
    step_size = 500
    generation_count = 2
    print "Creating initial population"
    while (True):
        if (last_generation <> generator.getCurrentGenerationNumber()):
            print "Size", len(generator.getCurrentPopulation().getIndividuals())
            print "Best", generator.getGenerationBest(last_generation)
            last_generation = generator.getCurrentGenerationNumber()
            if (last_generation > generation_count):
                break
            print "\nGeneration ", last_generation
        fitnesses = generator.runIterativeEvolution(step_size)
        if (len(fitnesses) > step_size):
            print "Fitness mismatch. Got", len(fitnesses)
