"""
FIXMEs:
    - Move scales/offsets/thresholds to constructor
"""
from math import *

from qt import *
from qtgl import *
from OpenGL.GL import *
from OpenGL.GLU import *
from copy import *

from GP_Interface import *

class GenerationsGraph(QGLWidget):
    def __init__(self, document=None, parent=None, *args):
        self._document = document
        apply(QGLWidget.__init__,(self,parent) + args)
        self._posX = 0.0
        self._posY = 0.0
        self._posZ = 0.0
        self._rotY = 45.0 * pi / 180.0
        self._rotXZ = 20.0 * pi / 180.0
        self._zoom = 11.0
        self._dot_height_offset = 0.005
        self._aspect = 1.0
        self._tracking_mouse = None
        self.resetContent(self._document.population_size, self._document.generation_count + 1)
        
        self.connect(self._document, PYSIGNAL("sigAddFitnesses"), self.addPoints)
        
        old_depth_list = []
        if (self._document._main_generation_graph <> None):
            old_depth_list = self._document._main_generation_graph._depth_list
        elif (len(self._document._loaded_depth_list) > 0):
            old_depth_list = self._document._loaded_depth_list
        
        if (len(old_depth_list) > 0):
            self._depth_list = copy(old_depth_list)
            if (len(self._depth_list) > self._depth):
                self._depth = len(self._depth_list)
            self._row_list = self._depth_list[-1]
            self.updateGL()
    def resetContent(self, width=0, depth=0):
        self._width = width
        self._depth = depth
        self._row_list = None
        self._depth_list = []
        if ((width == 0) or (depth == 0)):
            return
        self.GRID_SPACING_X = 0.025
        self.GRID_SPACING_Z = 0.25
        self.GRID_SPACING_X = self.GRID_SPACING_X * 500.0 / width
        self._offsetX = self.GRID_SPACING_X * float(width) / -2.0
        self._offsetZ = self.GRID_SPACING_Z * float(depth) / 2.0
        self.calculateCamera()
    def sizeHint(self):
        return QSize(600,250)
    def initializeGL(self):
        self.qglClearColor(QColor(0,0,0))
        glEnable(GL_DEPTH_TEST)
        glShadeModel(GL_SMOOTH)
    def calculateCamera(self):
        self.makeCurrent()  # INFO: Force this
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(45.0,self._aspect,1.0,20.0)
        self._fY = sin(self._rotXZ)
        ray = cos(self._rotXZ)
        self._fX = sin(self._rotY) * ray
        self._fZ = cos(self._rotY) * ray
        gluLookAt(self._posX + self._fX * self._zoom,self._posY + self._fY * self._zoom,self._posZ + self._fZ * self._zoom,self._posX,self._posY,self._posZ,0.0,1.0,0.0)
        glMatrixMode(GL_MODELVIEW)
    def mousePressEvent(self, eventMouse):
        self._tracking_mouse = eventMouse.button()
        self._mouse_start_pos = (eventMouse.x(),eventMouse.y())
        self.grabMouse()
    def mouseReleaseEvent(self, eventMouse):
        self.releaseMouse()
        self._tracking_mouse = None
    def mouseMoveEvent(self, eventMouse):
        if (self._tracking_mouse == None):
            return
        pos_diff = (eventMouse.x() - self._mouse_start_pos[0], self._mouse_start_pos[1] - eventMouse.y())
        self._mouse_start_pos = (eventMouse.x(),eventMouse.y())
        divider = 100.0
        if (self._tracking_mouse == Qt.LeftButton):
            self._rotY -= float(pos_diff[0]) / divider
            self._rotXZ -= float(pos_diff[1]) / divider
        elif (self._tracking_mouse == Qt.RightButton):
            self._zoom -= float(pos_diff[1]) / 50.0
        elif (self._tracking_mouse == Qt.MidButton):
            self._posX += float(pos_diff[0]) / divider * self._fZ
            self._posZ -= float(pos_diff[0]) / divider * self._fX
            self._posX -= float(pos_diff[1]) / divider * self._fX
            self._posZ -= float(pos_diff[1]) / divider * self._fZ
        self.calculateCamera()
        self.updateGL()
    def wheelEvent(self, eventWheel):
        self._zoom += eventWheel.delta() / 240.0
        self.calculateCamera()
        self.updateGL()
    def resizeGL(self,width,height):
        glViewport(0,0,width,height)
        self._aspect = 1.0
        if height > 0:
            self._aspect = float(width)/float(height)
        self.calculateCamera()
    def paintGL(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        if ((self._width == 0) or (self._depth == 0)):
            return
        glPushMatrix()
        glRotatef(self._rotY,0.0,1.0,0.0)
        glPushMatrix()
        glTranslate(self._offsetX, 0.0, 2.0 * self._offsetZ)
        glColor3f(0.1,0.1,0.1)
        glBegin(GL_QUADS)
        glVertex(0.0,0.0,0.0)
        glVertex(0.0, 0.0, self._depth * -self.GRID_SPACING_Z)
        glVertex((self._width - 1.0) * self.GRID_SPACING_X, 0.0, self._depth * -self.GRID_SPACING_Z)
        glVertex((self._width - 1.0) * self.GRID_SPACING_X, 0.0, 0.0)
        glEnd()
        for z in range(self._depth):
            fZ = float(z) * -self.GRID_SPACING_Z
            if (z < len(self._depth_list)):
                row_list = self._depth_list[z]
                next_row_list = []
                if ((z + 1) < len(self._depth_list)):
                    next_row_list = self._depth_list[z + 1]
                prev_row_list = []
                if (z > 0):
                    prev_row_list = self._depth_list[z - 1]
                for x in range(self._width - 1):
                    if (x < len(row_list)):
                        if (x > (self._width - 1)):
                            break
                        fX = float(x) * self.GRID_SPACING_X

                        colourScale = 200.0
                        heightScale = 64.0

                        fY = float(row_list[x]) / heightScale
                        fColour = float(row_list[x]) / colourScale

                        fAlongY = fY
                        fAlongColour = fColour
                        if ((x + 1) < len(row_list)):
                            fAlongY = float(row_list[x+1]) / heightScale
                            fAlongColour = float(row_list[x+1]) / colourScale
                        elif ((z > 0) and ((x + 1) < len(prev_row_list))):
                            fAlongY = float(prev_row_list[x+1]) / heightScale
                            fAlongColour = float(prev_row_list[x+1]) / colourScale

                        fNextY = fY
                        fNextColour = fColour
                        if (x < len(next_row_list)):
                            fNextY = float(next_row_list[x]) / heightScale
                            fNextColour = float(next_row_list[x]) / colourScale

                        fNextAlongY = fAlongY
                        fNextAlongColour = fAlongColour
                        if ((x + 1) < len(next_row_list)):
                            fNextAlongY = float(next_row_list[x+1]) / heightScale
                            fNextAlongColour = float(next_row_list[x+1]) / colourScale

                        colourBase1 = float(z) / float(self._depth) / 5.0
                        colourOffset = 0.3
                        threshold = 0.4
                        colourBase2 = 0.0
                        if ((fColour > threshold) or (fNextColour > threshold) or (fNextAlongColour > threshold) or (fAlongColour > threshold)):
                            colourBase1 = float(z) / float(self._depth) / 2.0
#                           colourBase2 = 0.3

                        glBegin(GL_QUADS)
                        glColor3f(colourOffset + fColour,colourBase2,colourBase1)
                        glVertex(fX, fY, fZ)
                        glColor3f(colourOffset + fNextColour,colourBase2,colourBase1)
                        glVertex(fX, fNextY, fZ - self.GRID_SPACING_Z)
                        glColor3f(colourOffset + fNextAlongColour,colourBase2,colourBase1)
                        glVertex(fX + self.GRID_SPACING_X, fNextAlongY, fZ - self.GRID_SPACING_Z)
                        glColor3f(colourOffset + fAlongColour,colourBase2,colourBase1)
                        glVertex(fX + self.GRID_SPACING_X, fAlongY, fZ)
                        glEnd()
                    else:
                        glColor3f(0.9,0.9,0.9)
                        glBegin(GL_POINTS)
                        fX = float(x) * self.GRID_SPACING_X
                        glVertex(fX, self._dot_height_offset, fZ)
                        glEnd()
            else:
                glColor3f(0.5,0.5,0.5)
                for x in range(self._width):
                    glBegin(GL_POINTS)
                    fX = float(x) * self.GRID_SPACING_X
                    glVertex(fX, self._dot_height_offset, fZ)
                    glEnd()
        glPopMatrix()
        glPopMatrix()
    def addPoints(self, point_list):
        if (self._row_list == None):
            self._depth_list.append([])
            self._row_list = self._depth_list[-1]
        if (len(self._depth_list) > self._depth):
            self._depth = len(self._depth_list)
        self._row_list.extend(point_list)
        if (len(self._row_list) == self._width):
            self._row_list = None
        self.updateGL()

if __name__ == "__main__":
    import sys
    app = QApplication(sys.argv)
    population_size = 500
    generation_count = 10
    reproduction_count = population_size / 10
    crossover_count = population_size - reproduction_count
    step_size = population_size / 5
    generator = IterativeGeneration(population_size, reproduction_count, crossover_count)
    graph = GenerationsGraph()
    
    graph.resetContent(population_size, (generation_count + 1))

    app.setMainWidget(graph)
    graph.show()
    graph.updateGL()

    app.processEvents()

    while (generator.getCurrentGenerationNumber() <= generation_count):
        fitnesses = generator.runIterativeEvolution(step_size)
#       if (len(fitnesses) > step_size):
#           print "Fitness list length problem", len(fitnesses)
#           break
        graph.addPoints(fitnesses)
        graph.updateGL()
        app.processEvents()

#   app.exec_loop()
