"""
FIXMEs:
    - Jump to (0,0) on new tile click
"""
from math import *
import PIL.Image
from qt import *
from qtgl import *
from OpenGL.GL import *
from OpenGL.GLU import *
from ThreeD import *

from GP_Interface import *

# glSelectWithCallback utility
def mySelectWithCallback(x, y, callback, xsize = 5, ysize = 5, buffer_size = 512):
    '''glSelectWithCallback(int x, int y, Callable callback, int xsize=5, int ysize=5)
  x,y -- x and y window coordinates for the center of the pick box
  rendercallback -- callback (callable Python object) taking 0 arguments
    which performs pick-mode rendering
  xsize,ysize -- x and y dimensions of the pick box (default = 5x5)
The function returns a tuple (possibly empty) of:
  ( (minimumzdepth, maximumzdepth, (name, name, name),...)
    minimumzdepth, maximumzdepth -- depths in integer format
      If you want the more traditional 0.0 to 1.0 numbers, divide
      by (2**32)-1
      If you want the physical depth, multiply that by the frustrum depth and
        add your near clipping plane.
    name -- the names (integers) used in calls to glPushName( int )'''
    # this import needs to be late binding or Python gets stuck in an infinite import loop
    from OpenGL.GLU import gluPickMatrix
    viewport = glGetIntegerv(GL_VIEWPORT)
    glSelectBuffer(buffer_size)
    glRenderMode(GL_SELECT)
    glInitNames()
    glMatrixMode(GL_PROJECTION)
    previousviewmatrix = glGetDoublev(GL_PROJECTION_MATRIX)
    glPushMatrix()
    glLoadIdentity()
    gluPickMatrix(x, viewport[3] - y, xsize, ysize, viewport)
    glMultMatrixd(previousviewmatrix)
    callback()
    glFlush()
#   glMatrixMode(GL_PROJECTION)
    glPopMatrix()
    glMatrixMode(GL_MODELVIEW)
#   glLoadMatrixd(previousviewmatrix)
    return glRenderMode(GL_RENDER)
#   return ()

class RoomView(QGLWidget):
    def __init__(self, document=None, parent=None, *args):
        self._document = document
        format = QGLFormat()
        format.setAlpha(True)
#       format.setStencil(True)
        apply(QGLWidget.__init__,(self,format, parent) + args)
        self._hintSize = QSize(500,500)
        self.setMouseTracking(True)
        self._posX = 0.0
        self._posY = 0.0
        self._posZ = 0.0
        self._viewvectX = 0.0
        self._viewvectY = 0.0
        self._viewvectZ = 0.0
        self._rotY = 45.0 * pi / 180.0
        self._rotXZ = 45.0 * pi / 180.0
        self._min_rotXZ = 2.0 * pi / 180.0
        self._max_rotXZ = 89.0 * pi / 180.0
        self._zoom = 9.0
        self._max_zoom = 9.6
        self._aspect = 1.0
        self._tile_height_offset = 0.01
        self._tracking_mouse = None
        self._tick_interval = 40
        self._highlight_tile = None
        self._animation_timer = -1
        self._particle_timer = -1
        self._update_partner = None
        self.resetRoom()
        
        if (document._viewing_individual <> None):
            program = document._individual_list[document._viewing_individual].getProgram()
            robot = Robot()
            if (document._robot_position <> None):
                robot = document._robot_position
            self.setProgram(program, robot)
    def loadResources(self):
#       self._image = TextureImage("rainbow.bmp")
#       self._image_brick = TextureImage("brick.bmp")
        self._image_marble = TextureImage("marble.bmp")
#       self._image_sky = TextureImage("sky.tga")
        self._image_billboard = TextureImage("opacity.tga")
        self._particle_system = ParticleEmitter(self._image_billboard, (0.8,0.5,0.0,0.5))
        import character
        self._model_robot = character.character(True)
        import sky
        self._model_sky = sky.sky(True)
    def sizeHint(self):
        return self._hintSize
    def closeEvent(self, eventClose):
        self.killTimers()
        QWidget.closeEvent(self, eventClose)
    def setUpdatePartner(self, partner):
        self._update_partner = partner
    def initializeGL(self):
        self.qglClearColor(QColor(0,0,0))
        glEnable(GL_DEPTH_TEST)
        glShadeModel(GL_SMOOTH)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glEnable(GL_COLOR_MATERIAL)
        glLight(GL_LIGHT0, GL_AMBIENT, (0.3,0.3,0.3,1.0))
        glLight(GL_LIGHT0, GL_DIFFUSE, (0.7,0.7,0.7,1.0))
        glLight(GL_LIGHT0, GL_POSITION, (0.0,3.0,0.0,1.0))
        glEnable(GL_LIGHT0)
#       glPixelStorei(GL_UNPACK_ALIGNMENT, 4)
#       glPixelStorei(GL_UNPACK_ROW_LENGTH, 0)
#       glPixelStorei(GL_UNPACK_SKIP_ROWS, 0)
#       glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0)
#       glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
#       glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        self._sphere = gluNewQuadric()
        gluQuadricDrawStyle(self._sphere, GLU_LINE)
        self.loadResources()
    def calculateCamera(self):
        self.makeCurrent()  # INFO: Force this
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(45.0,self._aspect,0.2,22.3)
        self._viewvectY = sin(self._rotXZ)
        ray = cos(self._rotXZ)
        self._viewvectX = sin(self._rotY) * ray
        self._viewvectZ = cos(self._rotY) * ray
        self._eyeX = self._posX + self._viewvectX * self._zoom
        self._eyeY = self._posY + self._viewvectY * self._zoom
        self._eyeZ = self._posZ + self._viewvectZ * self._zoom
        gluLookAt(self._eyeX,self._eyeY,self._eyeZ,self._posX,self._posY,self._posZ,0.0,1.0,0.0)
        glMatrixMode(GL_MODELVIEW)
    def mousePressEvent(self, eventMouse):
        if (eventMouse.state() == Qt.ShiftButton):
            if (self._highlight_tile and (self._robot > 0)):
                self.setRobot(Robot(self._highlight_tile[0] + 1, self._highlight_tile[1] + 1))
                self.updateGL()
                if (self._rate > 0.0):
                    self.toggletimerAnimation()
            return
        self._tracking_mouse = eventMouse.button()
        self._mouse_start_pos = (eventMouse.x(),eventMouse.y())
        self.grabMouse()
    def mouseReleaseEvent(self, eventMouse):
        self.releaseMouse()
        self._tracking_mouse = None
    def mouseMoveEvent(self, eventMouse):
        if not self._tracking_mouse:
            if (eventMouse.state() == Qt.ShiftButton):
                self.hittestTile(eventMouse.x(),eventMouse.y())
                self.updateGL()
            elif (self._highlight_tile):
                self._highlight_tile = None
            return
        pos_diff = (eventMouse.x() - self._mouse_start_pos[0], self._mouse_start_pos[1] - eventMouse.y())
        self._mouse_start_pos = (eventMouse.x(),eventMouse.y())
        divider = 100.0
        if (self._tracking_mouse == Qt.LeftButton):
            self._rotY -= float(pos_diff[0]) / divider
            self._rotXZ -= float(pos_diff[1]) / divider
            if (self._rotXZ < self._min_rotXZ):
                self._rotXZ = self._min_rotXZ
            elif (self._rotXZ > self._max_rotXZ):
                self._rotXZ = self._max_rotXZ
        elif (self._tracking_mouse == Qt.RightButton):
            self._zoom -= float(pos_diff[1]) / 50.0
            if (self._zoom > self._max_zoom):
                self._zoom = self._max_zoom
            elif (self._zoom < 0.0):
                self._zoom = 0.0
        elif (self._tracking_mouse == Qt.MidButton):
            self._posX += float(pos_diff[0]) / divider * self._viewvectZ
            self._posZ -= float(pos_diff[0]) / divider * self._viewvectX
            self._posX -= float(pos_diff[1]) / divider * self._viewvectX
            self._posZ -= float(pos_diff[1]) / divider * self._viewvectZ
        self.calculateCamera()
        self.updateGL()
    def wheelEvent(self, eventWheel):
        self._zoom += eventWheel.delta() / 240.0
        if (self._zoom > self._max_zoom):
            self._zoom = self._max_zoom
        elif (self._zoom < 0.0):
            self._zoom = 0.0
        self.calculateCamera()
        self.updateGL()
    def resizeGL(self,width,height):
        glViewport(0,0,width,height)
        self._aspect = 1.0
        if height > 0:
            self._aspect = float(width)/float(height)
        self.calculateCamera()
    ##################
    def resetRoom(self):
        self._program = None
        self.resetRobot()
        self._rate = 0.0
        self._old_rate = 0.0
        self._robot_facing = 0.0
    def resetRobot(self):
        self._robot = 0
        self._path_complete = 0.0
        self._robot_list = []
        self._path_complete = 0.0
        self._dead_end = False
#        self._robot_facing = 0.0
    def setRobot(self, robot):
        self.resetRobot()
        self._robot_list = [robot]
        self._robot = 1
        self._trail_offset_height = 0.1
        if (self._particle_timer > -1):
            print "Stopping particle system..."
            self._particle_system.stop()
        if (self._update_partner):
            self._update_partner.updateRobot(robot)
    def setProgram(self, program=None, robot=None):
        self._program = program
        if robot:
            self.setRobot(robot)
        elif (len(self._robot_list) > 0):
            first_robot = self._robot_list[0]
            self.setRobot(first_robot)
        self.updateGL()
    def setRate(self, rate=0.0):
        if (self._robot == 0):
            return
        if (self._rate <> 0.0):
            self._old_rate = self._rate
        if (self._old_rate == 0.0):
            self._old_rate = rate
        self._rate = rate
        print "New rate =", self._rate, ", Old rate =", self._old_rate
        if ((self._rate < 0.0) and (self._robot <= 1)):
            print "Can't go any further back than starting position"
            return
        if ((self._rate > 0.0) and (self._robot == len(self._robot_list)) and (self._dead_end == True)):
            print "Can't go further than a dead end"
            return
        if (self._rate <> 0.0):
#            if (((self._rate < 0.0) and (self._old_rate > 0.0)) or ((self._rate > 0.0) and (self._rate < 0.0))):
#                self._path_complete = 1.0 - self._path_complete
#                print "Path swapped"
            if ((self._rate < 0.0) and (self._robot == len(self._robot_list)) and (self._dead_end == True)):
                print "Stopping particle system..."
                self._particle_system.stop()
            self.toggletimerAnimation()
        else:
            self.toggletimerAnimation(False)
    def drawRobot(self, mirror=False):
        glPushMatrix()
        glColor(0.8,0.0,0.0)
        (fX,fZ) = self.getRobotPosition()
        glTranslate(fX,0.0,fZ)
        fRot = self.getRobotRotation()
        glRotate(fRot,0.0,1.0,0.0)
        glEnable(GL_LIGHTING)
#        if (self._dead_end == True):
#            self._model_robot.render()
#        else:
        self._model_robot.renderObject("Torso")
        self._model_robot.renderObject("Head")
        self._model_robot.renderObject("ShoulderL")
        self._model_robot.renderObject("ShoulderR")
        rotL = sin(self._path_complete * pi * 2.0) * 30.0 - 20.0
        rotR = cos(self._path_complete * pi * 2.0) * 30.0 - 20.0
        self._model_robot.renderObject("ArmL",(rotL,0.0,0.0))
        self._model_robot.renderObject("ArmR",(rotR,0.0,0.0))
        self._model_robot.renderObject("LegL",(rotR,0.0,0.0))
        self._model_robot.renderObject("LegR",(rotL,0.0,0.0))
        glDisable(GL_LIGHTING)
#       gluSphere(self._sphere, 0.25, 10, 10)
        glPopMatrix()
    def drawTrail(self):
        if (len(self._robot_list) <= 1):
            return
        step = 1.0 / float(len(self._robot_list))
        trail_colour = 0.0
        glColor(0.0,0.0,0.0)
        glPushMatrix()
        glBegin(GL_LINES)
        fY = self._trail_offset_height
        first_robot = self._robot_list[0]
        (fX1,fZ1) = self.getRobotPosition(False, first_robot)
        subtraction = 2
        if (self._dead_end == True):
            subtraction = 1
        for i in range(len(self._robot_list) - subtraction):
            (fX2,fZ2) = self.getRobotPosition(False, self._robot_list[i + 1])
            glColor(trail_colour,0.0,0.0)
            glVertex(fX1,fY,fZ1)
            trail_colour += step
            glColor(trail_colour,0.0,0.0)
            glVertex(fX2,fY,fZ2)
            fX1, fZ1 = fX2, fZ2
            fY += 0.0005
        if ((self._robot == (len(self._robot_list) - 1)) and (self._path_complete > 0.0) and (self._dead_end == False)):
            trail_colour += step
            (fX2,fZ2) = self.getRobotPosition()
            glVertex(fX1,fY,fZ1)
            glColor(trail_colour,0.0,0.0)
            glVertex(fX2,fY,fZ2)
        glEnd()
        glPopMatrix()
    def getTileUnderRobot(self):    # Zero-based
        (fX,fZ) = self.getRobotPosition()
        x = int(fX)
        z = int(fZ)
        return (x,z)
    def getRobotRotation(self):
        if ((len(self._robot_list) <= 1) or (self._robot == len(self._robot_list))):
            return self._robot_facing
        current_robot = self._robot_list[self._robot - 1]
        next_robot = self.getNextRobot()
        x = next_robot.getX() - current_robot.getX()
        if (x <> 0):
            self._robot_facing = float(x) * 90.0
        else:
            y = next_robot.getY() - current_robot.getY()
            if (y < 0):
                self._robot_facing = 180.0
            else:
                self._robot_facing = 0.0
        return self._robot_facing
    def getRobotPosition(self,calc_path=True,robot=None):
        current_robot = robot
        if not current_robot:
            if (self._robot > 0):
                if (self._robot <= len(self._robot_list)):
                    current_robot = self._robot_list[self._robot - 1]
            else:
                return (0.0,0.0)
        if (not current_robot):
            return (0.0,0.0)
        fX = float((current_robot.getX() - 1)) + 0.5
        fZ = float((current_robot.getY() - 1)) + 0.5
        if ((calc_path == True) and (self._path_complete > 0.0) and (len(self._robot_list) > 1)):
            next_robot = self.getNextRobot()
            fX += self._path_complete * float(next_robot.getX() - current_robot.getX())
            fZ += self._path_complete * float(next_robot.getY() - current_robot.getY())
        return (fX,fZ)
    def getNextRobot(self):
        if (self._rate == 0.0):
            return  self._robot_list[self._robot]
        step = 1
        if (self._rate < 0.0):
            step = -1
        try:
            return self._robot_list[(self._robot - 1) + step]
        except:
            return self._robot_list[0]
    def toggletimerAnimation(self, enable=True):
        if ((enable == True) and (self._animation_timer == -1)):
            self._animation_timer = self.startTimer(self._tick_interval)
            print "ANIMATION TIMER ON"
        elif ((enable == False) and (self._animation_timer > -1)):
            self.killTimer(self._animation_timer)
            print "ANIMATION TIMER OFF"
            self._animation_timer = -1
    def toggletimerParticleSystem(self, enable=True):
        if ((enable == True) and (self._particle_timer == -1)):
            self._particle_system.reset()
            self._particle_timer = self.startTimer(self._tick_interval)
            print "PARTICLE SYSTEM TIMER ON"
        elif (enable == True): # INFO: Timer's already running, but system probably stopping
            self._particle_system.start()
            print "Restarting particle system..."
        elif ((enable == False) and (self._particle_timer > -1)):
            self.killTimer(self._particle_timer)
            print "PARTICLE SYSTEM TIMER OFF"
            self._particle_timer = -1
    def timerEvent(self, timerEvent):
        if (timerEvent.timerId() == self._animation_timer):
            if (len(self._robot_list) == 0) or (not self._program):
                return
            if ((self._robot == 1) and (self._rate < 0.0)):
                self.toggletimerAnimation(False)
                self._rate = 0.0
                return
            if (self._path_complete < 1.0):
                self._path_complete += abs(self._rate)
            if (self._path_complete >= 1.0):
                self._path_complete = 0.0
                if (self._rate > 0.0):
                    self._robot += 1
                    if ((self._dead_end == True) and (self._robot == len(self._robot_list))):
                        print "Got back to dead end"
                        self.toggletimerAnimation(False)
                        self.toggletimerParticleSystem()
                else:
                    self._robot -= 1
                    if (self._robot == 1):
                        print "REACHED BEGINNING"
                        self.toggletimerAnimation(False)
                if (self._update_partner):
                    self._update_partner.updateRobot(self._robot_list[self._robot - 1])
            if (self._robot == len(self._robot_list) and (self._dead_end == False)):
                current_robot = self._robot_list[self._robot - 1]
                next_robot = Robot(current_robot.getX(), current_robot.getY())
                next_robot.move(self._program)
#               if (len(self._robot_list) > 5):
#                   pass
                if (next_robot == current_robot):
                    self._dead_end = True
                    print "SET DEAD END"
                    self.toggletimerAnimation(False)
                    self.toggletimerParticleSystem()
                else:
                    self._robot_list.append(next_robot)
            self.updateGL()
        elif (timerEvent.timerId() == self._particle_timer):
            if (self._particle_system.tick() == False):
                self.toggletimerParticleSystem(False)
            self.updateGL()
    ##################
    def drawRoom(self, mirror=False):
        self._model_sky.render()
        if (self._robot > 0):
            glPushMatrix()
            glTranslate(-4.0,0.0,-4.0)
            self.drawTrail()
            self.drawRobot(mirror)
            glPopMatrix()
            if (self._particle_timer > -1):
                (fX,fZ) = self.getRobotPosition()
                fX -= 4.0
                fZ -= 4.0
                offset = 0.32
                if (self._robot_facing == 0.0):
                    fZ += offset
                elif (self._robot_facing == 90.0):
                    fX += offset
                elif (self._robot_facing == 180.0):
                    fZ -= offset
                else:
                    fX -= offset
                emitter_height = 1.0
                self._particle_system.move((fX,emitter_height,fZ))
                self._particle_system.render((self._eyeX,self._eyeY,self._eyeZ), mirror)
        glPushMatrix()
#       glColor(0.5,0.5,0.5)
#       glDrawPixels(128,128,GL_RGBA,GL_UNSIGNED_BYTE,self._image)
#       glEnable(GL_TEXTURE_2D)
#       glTexImage2D(GL_TEXTURE_2D, 0, 3, self._image.width(), self._image.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, self._image.bits())
#       glBegin(GL_QUADS)
#       glTexCoord(0.0,0.0)
#       glVertex(-3.0,0.0,3.0)
#       glTexCoord(1.0,0.0)
#       glVertex(3.0,0.0,-3.0)
#       glTexCoord(1.0,1.0)
#       glVertex(3.0,3.0,-3.0)
#       glTexCoord(0.0,1.0)
#       glVertex(-3.0,3.0,3.0)
#       glEnd()
#       glDisable(GL_TEXTURE_2D)
        glPopMatrix()
    def hittestTileRender(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        self.drawFloor()
    def hittestTile(self,x,y):
#       buffer = glSelectWithCallback(x,y,self.hittestTileRender)
        buffer = mySelectWithCallback(x,y,self.hittestTileRender)
        if len(buffer) > 1:
            name1 = buffer[1]
            tile = name1[2][0] - 1
            if ((tile >= 0) and (tile < 64)):
                self._highlight_tile = (tile % 8,tile / 8)
                return
        self._highlight_tile = None
    def drawFloor(self):
        glPushMatrix()
        glTranslate(-4.0,0.0,-4.0)
        glEnable(GL_BLEND)
        glColor(0.3,0.3,0.3,0.3)
        floor_size = 20.0
#       glEnable(GL_TEXTURE_2D)
#       glTexImage2D(GL_TEXTURE_2D, 0, 3, self._image_brick.width(), self._image_brick.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, self._image_brick.bits())
        glBegin(GL_QUADS)
        glTexCoord(0.0,0.0)
        glVertex3f(-floor_size,0.0,-floor_size)
        glTexCoord(1.0,0.0)
        glVertex3f(-floor_size,0.0,floor_size)
        glTexCoord(1.0,1.0)
        glVertex3f(floor_size,0.0,floor_size)
        glTexCoord(0.0,1.0)
        glVertex3f(floor_size,0.0,-floor_size)
        glEnd()
#       glDisable(GL_TEXTURE_2D)
        glEnable(GL_TEXTURE_2D)
        glPushName(0)
        fY = self._tile_height_offset
        border = 0.035
        tileX = -1
        tileZ = -1
        if (self._robot > 0):
            (tileX,tileZ) = self.getTileUnderRobot()
        for z in range(8):
            for x in range(8):
                if (((x == 3) or (x == 4)) and ((z == 6) or (z == 7))):
                    continue
                fX = float(x)
                fZ = float(z)
                colorR = 0.6
                colorG = 0.6
                colorB = 0.6
                colorA = 0.5
                if ((x == tileX) and (z == tileZ)):
                    colorA = 0.9
                    if (self._dead_end == True):
                        colorG = 0.9
                    else:
                        colorR = 0.9
                if (self._highlight_tile):
                    if ((x == self._highlight_tile[0]) and (z == self._highlight_tile[1])):
                        colorB = 0.9
                        colorA = 0.9
                glColor(colorR,colorG,colorB,colorA)
                glLoadName((z * 8) + x + 1)
                glTexImage2D(GL_TEXTURE_2D, 0, 3, self._image_marble.width(), self._image_marble.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, self._image_marble.bits())
                glBegin(GL_QUADS)
                glTexCoord(0.0,0.0)
                glVertex3f(fX + border, fY, fZ + border)
                glTexCoord(1.0,0.0)
                glVertex3f(fX - border + 1.0, fY, fZ + border)
                glTexCoord(1.0,1.0)
                glVertex3f(fX - border + 1.0, fY, fZ - border + 1.0)
                glTexCoord(0.0,1.0)
                glVertex3f(fX + border, fY, fZ - border + 1.0)
                glEnd()
        glDisable(GL_TEXTURE_2D)
        glDisable(GL_BLEND)
        glPopMatrix()
    def paintGL(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glPushMatrix()
        glRotatef(self._rotY,0.0,1.0,0.0)
#       glBegin(GL_LINES)
#       glColor3f(0.0,0.0,0.0)
#       glVertex3f(0.0,0.0,0.0)
#       glColor3f(1.0,0.0,0.0)
#       glVertex3f(5.0,0.0,0.0)
#       glColor3f(0.0,0.0,0.0)
#       glVertex3f(0.0,0.0,0.0)
#       glColor3f(0.0,1.0,0.0)
#       glVertex3f(0.0,5.0,0.0)
#       glColor3f(0.0,0.0,0.0)
#       glVertex3f(0.0,0.0,0.0)
#       glColor3f(0.0,0.0,1.0)
#       glVertex3f(0.0,0.0,5.0)
#       glEnd()
#       glEnable(GL_STENCIL_TEST)
#       glClearStencil()
#       glStencilFunc(GL_ALWAYS, 1, 1)
#       glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE)
#       self.drawFloor()
#       glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
#       glStencilFunc(GL_EQUAL, 1, 1)
#       glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP)
        glPushMatrix()
        glScale(1.0,-1.0,1.0)
        self.drawRoom(True)
        glPopMatrix()
#       glDisable(GL_STENCIL_TEST)
        self.drawFloor()
        self.drawRoom()
        glPopMatrix()

perfectProgram = If(And(Or(N(),NE()),Not(E())), East(), If( And(Or(E(),SE()),Not(S())), South(), If( And(Or(S(),SW()),Not(W())), West(), North())))
sample1 = If(And(One(),N()),Or(South(),E()),One())
sample2 = And(Or(N(),East()),South())

if __name__ == "__main__":
    import sys
    app = QApplication(sys.argv)
    view = RoomView()

    app.setMainWidget(view)
    view.show()
    view.updateGL()

    view.setProgram(perfectProgram, Robot())
    view.setRate(0.1)

    app.exec_loop()
