from random import *
from math import *
from PIL import Image
from OpenGL.GL import *
from OpenGL.GLU import *

MAP_DIFFUSE = 0
MAP_OPACITY = 1

class TextureImage:
    def __init__(self, path=None):
        self._image = None
        self._bits = None
        if path:
            self.loadImage(path)
    def loadImage(self, path):
        self._image = Image.open(path)
        print "Loaded image:", path, "(", self.width(), "x", self.height(), ")"
        if self._image:
            mode = "RGBX"
            if (self._image.mode == "RGBA"):
                mode = "RGBA"
                print "Image has alpha layer"
            self._bits = self._image.tostring("raw", mode, 0, -1)
        else:
            print "FAILED TO LOAD IMAGE:", path
    def width(self):
        return self._image.size[0]
    def height(self):
        return self._image.size[1]
    def bits(self):
        return self._bits

class Billboard:
    def __init__(self, texture_image, size=1.0):
        self.texture = texture_image
        self.size = size
        self.posX = 0.0
        self.posY = 0.0
        self.posZ = 0.0
    def move(self, coords):
        self.posX = coords[0]
        self.posY = coords[1]
        self.posZ = coords[2]
    def moveOffset(self, coords):
        self.posX += coords[0]
        self.posY += coords[1]
        self.posZ += coords[2]
    def render(self, eye_coords, mirror=False):
        glPushAttrib(GL_ENABLE_BIT)
        glEnable(GL_TEXTURE_2D)
        glTexImage2D(GL_TEXTURE_2D, 0, 4, self.texture.width(), self.texture.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, self.texture.bits())
        glPushMatrix()
        glTranslate(self.posX,self.posY,self.posZ)
        diffZ = eye_coords[2] - self.posZ
        fY = 0.0
        fX = 0.0
        if (diffZ <> 0.0):
            diffX = eye_coords[0] - self.posX
            fY = atan(diffX/diffZ) * 180.0 / pi
            alpha = sqrt(diffX**2 + diffZ**2)
            if ((diffZ > 0.0) and (mirror == False)):
                alpha = -alpha
            if ((diffZ < 0.0) and (mirror == True)):
                alpha = -alpha
            if (alpha <> 0.0):
                diffY = eye_coords[1] - self.posY
                fX = atan(diffY/alpha) * 180 / pi
            else:
                pass
        else:
            fX = 90.0
        glRotate(fY,0.0,1.0,0.0)
        glRotate(fX,1.0,0.0,0.0)
        glBegin(GL_QUADS)
        glTexCoord(0.0,0.0)
        glVertex(-self.size,-self.size,0.0)
        glTexCoord(1.0,0.0)
        glVertex(self.size,-self.size,0.0)
        glTexCoord(1.0,1.0)
        glVertex(self.size,self.size,0.0)
        glTexCoord(0.0,1.0)
        glVertex(-self.size,self.size,0.0)
        glEnd()
        glPopMatrix()
        glDisable(GL_TEXTURE_2D)
        glPopAttrib()

class Particle(Billboard):
    def __init__(self, randomiser, *args):
        apply(Billboard.__init__,(self,) + args)
        self.randomiser = randomiser
        self.reset()
    def reset(self):
        self.age_value = 0
        self.rotY = float(self.randomiser.randint(0,359)) / 180 * pi
        self.rotXZ = float(self.randomiser.randint(45,85)) / 180 * pi
    def age(self):
        self.age_value += 1
    def getAge(self):
        return self.age_value
    def render(self, eye_coords, mirror):
        step = float(self.age_value) * 0.02
        fY = sin(self.rotXZ) * step
        ray = cos(self.rotXZ)
        fX = sin(self.rotY) * ray * step
        fZ = cos(self.rotY) * ray * step
        self.moveOffset((fX,fY,fZ))
        Billboard.render(self, eye_coords, mirror)

class ParticleEmitter:
    def __init__(self, texture_image, colour=(0.8,0.8,0.8,0.8), size=0.25, rate=1, lifetime=20, falloff=5):
        self.texture = texture_image
        self.size = size
        self.colour = colour
        self.rate = rate
        self.lifetime = lifetime
        self.falloff = falloff
        self.randomiser = Random()
        self.billboard_list = []
        for i in range(lifetime / rate):
            self.billboard_list.append(Particle(self.randomiser, self.texture, self.size))
        self.position = (0.0,0.0,0.0)
        self.reset()
    def reset(self):
        self.tick_count = 0
        self.next_particle = 0
        for particle in self.billboard_list:
            particle.reset()
        self.start()
    def start(self):
        self.emit = True
    def stop(self):
        self.emit = False
    def tick(self):
        ticked = self.emit
        for i in range(len(self.billboard_list)):
            particle = self.billboard_list[i]
            age = particle.getAge()
            if (age == 0):
                continue
            particle.age()
            ticked = True
            if (age > self.lifetime):
                particle.reset()
                self.next_particle = i
        if ((self.emit == True) and ((self.tick_count % self.rate) == 0)):
            self.billboard_list[self.next_particle].age()
            self.next_particle += 1
            if (self.next_particle == len(self.billboard_list)):
                self.next_particle = 0
        self.tick_count += 1
        return ticked
    def move(self, offset):
        self.position = offset
    def render(self, eye_coords, mirror=False):
        glPushAttrib(GL_COLOR_BUFFER_BIT)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE)
        glEnable(GL_ALPHA_TEST)
        glAlphaFunc(GL_GREATER, 0.3)
        for particle in self.billboard_list:
            alpha = self.colour[3]
            age = particle.getAge()
            if (age == 0):
                continue
            if (age > (self.lifetime - self.falloff)):
                alpha *= (float(self.lifetime - age) / float(self.falloff))
            glColor(self.colour[0],self.colour[1],self.colour[2],alpha)
            particle.move(self.position)
            particle.render(eye_coords, mirror)
        glDisable(GL_ALPHA_TEST)
        glDisable(GL_BLEND)
        glPopAttrib()

class Face:
    def __init__(self, vertices=None, edge_visibility=None, texture=None, normal=None, vertex_normals=None, vertex_colours=None, tex_coords=None):
        self.vertices = vertices
        self.vertex_normals = vertex_normals
        self.vertex_colours = vertex_colours
        self.normal = normal
        self.texture = texture
        self.edge_visibility = edge_visibility
        self.tex_coords = tex_coords
    def render(self):
        glBegin(GL_TRIANGLES)
        for i in range(3):
            if (self.vertex_normals):
                glNormal(self.vertex_normals[i][0],self.vertex_normals[i][1],self.vertex_normals[i][2])
            if (self.vertex_colours):
                glColor(self.vertex_colours[i][0],self.vertex_colours[i][1],self.vertex_colours[i][2])
            if (self.tex_coords):
                glTexCoord(self.tex_coords[i][0],self.tex_coords[i][1])
            glVertex(self.vertices[i][0],self.vertices[i][1],self.vertices[i][2])
        glEnd()

class Object:
    def __init__(self, wireframe_colour=None, offset=None, axial_rotation=None, angle=None, texture=None, faces=None):
        self.wireframe_colour = wireframe_colour
        self.offset = offset
        self.axial_rotation = axial_rotation
        self.angle = angle
        self.texture = texture
        self.display_list = None
        self.faces = faces
    def compile(self):
        self.display_list = glGenLists(1)
        print "Object display list:", self.display_list
        glNewList(self.display_list, GL_COMPILE)
        if (self.wireframe_colour):
            glColor(self.wireframe_colour[0],self.wireframe_colour[1],self.wireframe_colour[2])
#       glPushMatrix()
#       glTranslate(self.offset[0],self.offset[1],self.offset[2])
        if (self.texture):
            glEnable(GL_TEXTURE_2D)
            self.texture.useMap()
        else:
            glDisable(GL_TEXTURE_2D)
        for face in self.faces:
            face.render()
        if (self.texture):
            glDisable(GL_TEXTURE_2D)
#       glPopMatrix()
        glEndList()
    def render(self, rotation=None):
        glPushMatrix()
        glTranslate(self.offset[0],self.offset[1],self.offset[2])
        if (rotation):
            glRotate(rotation[0],1.0,0.0,0.0)
            glRotate(rotation[1],0.0,1.0,0.0)
            glRotate(rotation[2],0.0,0.0,1.0)
        glCallList(self.display_list)
        glPopMatrix()

class Texture:
    def __init__(self, map_list=None, ambient=None, diffuse=None, specular=None, shine=None, shine_strengh=None, transparency=None, self_illumination=None, sub_maps=None):
        self.map_list = map_list
        self.ambient = ambient
        self.diffuse = diffuse
        self.specular = specular
        self.shine = shine
        self.shine_strengh = shine_strengh
        self.transparency = transparency
        self.self_illumination = self_illumination
        self.sub_maps = sub_maps
        self.texture_images = {}
    def load(self):
        for i in range(len(self.map_list)):
            map_path = self.map_list[i]
            if (map_path):
                self.texture_images[i] = TextureImage(map_path)
#               print "Loaded image:", map_path
        for map in self.sub_maps:
            map.load()
    def useMap(self,map=MAP_DIFFUSE):
        map = self.texture_images[map]
        if (map):
            glTexImage2D(GL_TEXTURE_2D, 0, 3, map.width(), map.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, map.bits())

class Model:
    def __init__(self, load_now=False):
        # FIXME: Check attributes and set empty if non-existent
        self.display_list = None
        if (load_now == True):
            self.loadResources()
            self.compile()
    def loadResources(self):
        print "Loading resources for model:", self.__class__.__name__
        for texture in self.textures:
            print "Loading texture:", texture
            self.textures[texture].load()
    def compile(self):
        print "Compiling model:", self.__class__.__name__
        for object in self.objects:
            print "Compiling object:", object
            self.objects[object].compile()
        self.display_list = glGenLists(1)
        glNewList(self.display_list, GL_COMPILE)
        for object in self.objects:
            self.objects[object].render()
        glEndList()
        print "Model display list:", self.display_list, "\n"
    def render(self):
        glCallList(self.display_list)
    def renderObject(self, object, rotation=None):
        self.objects[object].render(rotation)

