"""
FUTURE THINGS TO DO:
-   Highlight the node that's actually being evaluated
"""

from qt import *

from GP_Imp import *

values = [One, Zero]
functions = [If, Or, And, Not]
actions = [North, East, South, West]
sensors = [N, NE, E, SE, S, SW, W, NW]

DEFAULT_ROUND_RECT  = 50

class TreeGraph(QWidget):
    def __init__(self, document=None, parent=None, *args):
        self._document = document
        apply(QWidget.__init__,(self, parent) + args)
        self.setEraseColor(QColor(255,255,255))
        self.height = 0
        self.width = 0
        self.current_tree = None
        self.is_robot_valid = False
        self.robot = None
        
        self.background_color = QColor(255,255,255)
        
        self.current_node_color = QColor(255, 192, 203)
        
        # self.node_color = QColor(248,248,255) #ghost white
        # self.node_color = QColor(255,250,250) #Snow
        self.node_color = QColor(255,255,240) # Ivory   
        
        self.value_node_color = QColor(255,250,173)
        self.function_node_color = QColor(173,180,255)
        self.action_node_color = QColor(187,255,173)
        
        self.sensor_node_color = QColor(255,206,206)
        self.active_sensor_node_color = QColor(255,99,71)
        
##~         self.current_node_color = QColor(255, 192, 203)
##~         self.value_node_color = QColor(100,149,237)
##~         self.function_node_color = QColor(50,205,50)
##~         self.action_node_color = QColor(255,215,0)
##~         self.sensor_node_color = QColor(135,206,250)
        
        self.node_width_size = 30
        self.node_height_size = 30
        
        self.extra_space = 15
        
        self.font_family =  'Courier' 
        self.font_size = 6 
        self.font_weight  = QFont.Normal
        self.font_italic = False
        
        self.margin = 25
        
        self.default_x = self.margin
        self.default_y = self.node_height_size
        
        self.x = self.margin
        self.y = self.node_height_size
        
        self.prev_x = self.x
        self.prev_y = self.y
        
        self.node_x = 0 
        self.node_y = 0
        
        self.first_node = 0
        
        self.width_space = 15
        self.height_space = 15
        
        self.pixmap = QPixmap(1,1)
        self.painter = QPainter(self.pixmap)
        
        self.connect(self._document, PYSIGNAL("sigViewIndividual"), self.showIndividual)
        self.connect(self._document, PYSIGNAL("sigUpdateRobot"), self.updateRobot)
        
        if (self._document._viewing_individual <> None):
            self.showIndividual(self._document._viewing_individual)
    
    def showIndividual(self, individual):
        try:
            individual = self._document._individual_list[individual]
            self.setNewTree(individual.getProgram())
        except:
            return
    
    def sizeHint(self):
        if ((self.width == 0) or (self.height == 0)):
            return QSize(300,400)
        return QSize(self.width, self.height)
    
    def paintTree(self):
        if (self.current_tree):
            if (self.pixmap.isNull() == True):
                self.pixmap = QPixmap(self.width,self.height)# INFO: Probably not required
            if (self.painter.isActive() == False):
                self.painter.begin(self)
            self.current_tree.resetData()
            self.drawProgramTree(self.current_tree)
            self.painter.end()
            self.x = self.default_x
            self.y = self.default_y
    
    def pixmapDraw(self):
        self.painter = QPainter(self.pixmap)
        self.pixmap.fill(self.background_color)
        self.paintTree()
        self.paintMe()
    
    def setNewTree(self, new_tree):
        self.current_tree = new_tree
        self.paintTree()
        self.height += self.margin
        self.setFixedSize(self.width,self.height)
        self.pixmap = QPixmap(self.width,self.height)
        self.pixmapDraw()
    
    def updateRobot(self, new_robot):
        self.robot = new_robot
        self.is_robot_valid = True
        self.pixmapDraw()
    
    def paintMe(self):
        bitBlt(self, 0, 0, self.pixmap)
    
    def paintEvent(self, ev):
        self.paintMe()
    
    def setParent(self,tree,parent=None):
        tree.parent = parent
        for child in tree.getChildren():
            self.setParent(child,tree)
    
    def drawProgramTree(self, tree):
        max_tree_height = tree.getTreeHeight()
        self.height = max_tree_height*self.node_height_size + (max_tree_height - 1)*self.height_space + self.margin
        self.setNodeLevels(tree, 1)
        self.setNodePositions(tree, max_tree_height)
        
        self.setParent(tree)
        self.adjustNodePositions(tree)
        
        self.drawTree(tree)
        self.drawBranches(tree)
    
    def setNodeLevels(self, tree, next_level):
        tree.setLevel(next_level)
        tree_children = tree.getChildren()
        if tree_children == []:
            return
        else:
            for child in tree_children:
                self.setNodeLevels(child, next_level+1)
    
    def getRightmost(self,tree,required_level,current_level):
        if (required_level == current_level):
            return tree
        children = tree.getChildren()
        if (len(children) > 0):
            retval = self.getRightmost(children[-1], required_level, current_level + 1)
            if (retval):
                return retval
            elif (len(children) >= 2):
                retval = self.getLeftChildOnSameLevel(children[-1], current_level, current_level)
                if (retval):
                    retval = self.getRightmost(retval, required_level, current_level+1)
                    if (retval):
                        return retval
                    elif (len(children) == 3):
                        retval = self.getLeftChildOnSameLevel(children[-2], current_level, current_level)
                        if (retval):
                            return self.getRightmost(retval, required_level, current_level +1)
        else:
            return None
    
    def getLeftChildOnSameLevel(self, tree, required_level, current_level):
        if (not tree.parent):
            return None
        i = 0
        for child in tree.parent.getChildren():
            if (tree is child):
                break
            i += 1
        if (i > 0):
            left = tree.parent.getChildren()[i - 1]
            if (required_level == current_level):
                return left
            right = self.getRightmost(left, required_level, current_level)
            if (right):
                return right
            else:
                if (i == 2):
                    left = tree.parent.getChildren()[i - 2]
                    right = self.getRightmost(left, required_level, current_level)
                    if (right):
                        return right
#                    else:
#                return self.getLeftChildOnSameLevel(tree.parent, required_level, current_level - 1)#self.getLeftChildOnSameLevel(left, required_level, current_level + 1)
        return self.getLeftChildOnSameLevel(tree.parent, required_level, current_level - 1)
    
    def adjustNodePositions(self, tree, level=1):
        left = self.getLeftChildOnSameLevel(tree,level,level)
        if left:
            node_x1 = left.getX() + self.node_width_size
            node_x2 = tree.getX()
            if node_x1 >= node_x2: #overlap
                x = node_x1 + self.extra_space
                tree.setX(x)
                self.resetWidth(x)
                for child in tree.getChildren():
                    self.moveTreeChildrenRight(child, x - node_x2)
        for child in tree.getChildren():
            self.adjustNodePositions(child, level + 1)
    
    def moveTreeChildrenRight(self, tree, diff):
        x = tree.getX() + diff
        tree.setX(x)
        self.resetWidth(x)
        for child in tree.getChildren():
            self.moveTreeChildrenRight(child, diff)
    
    def setNodePositions(self, tree, max_tree_height):
##~         print "self.x" , self.x, "self.y", self.y
        for level in range(max_tree_height, 0, -1):
            self.setNodePositionsHelper(tree, level, max_tree_height)
            self.y = self.y + self.node_height_size + self.height_space
            self.x = self.margin
    
    def setNodePositionsHelper(self, tree, level, max_tree_height):     
        if tree.getLevel() == level:
            if tree.getLevel() == max_tree_height:
                tree.setX(self.x)
                tree.setY(self.y)
                self.x = self.x + self.node_width_size + self.width_space
                self.y = self.y 
            else:
                tree_children = tree.getChildren()
                children_length = len(tree_children)
                if children_length == 0:         
                    if self.x <> self.margin:
                        tree.setX(self.x + self.node_width_size + self.width_space)
                    else:
                        tree.setX(self.x)
                    tree.setY(self.y)                    
                elif children_length == 1:
                    tree.setX(tree_children[0].getX())
                    tree.setY(tree_children[0].getY()+self.node_height_size+self.height_space)
                else:
                    far_right_child_x = tree_children[children_length-1].getX()
                    tmp_width_children = far_right_child_x + self.node_width_size - tree_children[0].getX()
                    x = far_right_child_x - (tmp_width_children/2) + (self.node_width_size/2)
                    y = tree_children[0].getY()+ self.node_height_size + self.height_space
                    tree.setX(x)
                    tree.setY(y)
                self.x = tree.getX()
                self.y = tree.getY()
            self.resetWidth(tree.getX())
            return
        else:
            for child in tree.getChildren():
                self.setNodePositionsHelper(child, level, max_tree_height)

    def resetWidth(self, new_tree_x):
        tmp_width = new_tree_x + self.node_width_size + self.margin
        if self.width < tmp_width:
            self.width = tmp_width
    
    def drawBranches(self, tree):
        tree_children = tree.getChildren()
        if tree_children == []:
            return
        else:
            xFrom = tree.getX() + (self.node_width_size/2)
            yFrom = self.height - tree.getY() + self.node_height_size
            for child in tree_children:
                xTo = child.getX() + (self.node_width_size/2)
                yTo = self.height - child.getY() 
                self.drawBranch(xFrom, yFrom, xTo, yTo)
                self.drawBranches(child)
    
    def drawBranch(self, xFrom, yFrom, xTo, yTo):
        self.painter.drawLine(xFrom,yFrom,xTo,yTo)
    
    def drawTree(self, tree):
        tree_children = tree.getChildren()
        if tree_children == []:
            self.drawNode(tree)
        else:
            self.drawNode(tree)
            for child in tree_children:
                self.drawTree(child)

    def drawNode(self, node):
        type = node.__class__
        x = node.getX()
        y = self.height - node.getY()
        w = self.node_width_size
        h = self.node_height_size
        
        if type in values:           # round rectange      
            self.painter.setBrush(QBrush(self.value_node_color))
            self.drawValue(x, y, w, h)
        elif type in functions:    # circle
            self.painter.setBrush(QBrush(self.function_node_color))            
            self.drawFunction(x, y, w, h)                
        elif type in actions:       # rectangle
            self.painter.setBrush(QBrush(self.action_node_color))                        
            self.drawAction(x, y, w, h)
        elif type in sensors:     # circle
            brush = QBrush(self.sensor_node_color)
            if self.is_robot_valid == True:
                if self.robot.getSensorValue(node) == 1:
                    brush = QBrush(QColor(self.active_sensor_node_color))
            self.painter.setBrush(brush)
            self.drawSensor(x, y, w, h)
        
        self.painter.setFont(QFont(self.font_family,self.font_size, self.font_weight, self.font_italic)) 
        self.painter.drawText(x, y, w, h,Qt.AlignCenter, node.__class__.__name__)

    def drawValue(self, x, y, w, h, xR = DEFAULT_ROUND_RECT, yR = DEFAULT_ROUND_RECT):
        self.painter.drawRoundRect(x,y,w,h,xR,yR)

    def drawFunction(self, x, y, w, h):
        self.painter.drawEllipse(x,y,w,h)

    def drawAction(self, x, y, w, h):
        self.painter.drawRect(x,y,w,h)

    def drawSensor(self, x, y, w, h):
        self.painter.drawEllipse(x,y,w,h)

generated4 = If(Or(NW(),North()),If(If(NE(),Or(East(),If(NW(),East(),North())),North()),If(If(NW(),Or(East(),North()),North()),North(),South()),South()),NE())
generated5 = And(And(Or(If(NE(),If(SE(),South(),East()),W()),If(And(And(NE(),Not(If(If(And(If(NE(),South(),East()),And(South(),And(N(),West()))),North(),East()),South(),East()))),West()),North(),West())),Or(North(),Zero())),Or(Or(Or(East(),S()),Zero()),North()))
generated6 = And(One(),If(N(),West(),If(North(),If(If(Or(NW(),North()),East(),And(W(),North())),If(If(Or(NW(),North()),East(),And(W(),North())),East(),E()),If(If(Or(NW(),North()),West(),E()),W(),Or(Or(NW(),North()),North()))),NW())))

if __name__ == "__main__":
    import sys
    app=QApplication(sys.argv)
    treeGraph = TreeGraph()
    treeGraph.setNewTree(generated6)
    app.setMainWidget(treeGraph)
    treeGraph.show()
    app.exec_loop()
