"""
mdidoc.py - document or application model.

copyright: (C) 2001, Boudewijn Rempt
email:     boud@rempt.xs4all.nl
"""
from qt import *
from resources import TRUE, FALSE
import cPickle
import math
from SettingsDialog import *
from GP_Interface import *

class MDIDoc(QObject):
    """
    The document represents the application model. The current
    document keeps a 'modified' state.

    signals: sigDocModified (boolean)
             sigDocTitleChanged (string)
             
    """
    def __init__(self, *args):
        apply(QObject.__init__, (self,)+args)
        self.newDocument()
        self._fileName=None
        self._title="Untitled"
        
        self._generated_generations = None
        self._viewing_generation = None
        self._viewing_individual = None
        self._individual_list = []
        self._robot_position = None
        self._main_generation_graph = None
        self._loaded_depth_list = []
        
        self.settings_dialog = None
        self.generations = None
        
        self._generation_control = None
        self._visualisation_control = None
    
    def addFitnesses(self, fitness_list):
        self.emit(PYSIGNAL("sigAddFitnesses"), (fitness_list,))
    
    def updateRobot(self, robot_position):
        self._robot_position = robot_position
        self.emit(PYSIGNAL("sigUpdateRobot"), (self._robot_position,))
    
    def viewGeneration(self, generation):
        self._viewing_generation = generation
        self.emit(PYSIGNAL("sigViewGeneration"), (self._viewing_generation,))
    
    def viewIndividual(self, individual):
        self._viewing_individual = individual
        if (self._viewing_individual <> None):
            self.emit(PYSIGNAL("sigViewIndividual"), (self._viewing_individual,))
    
    def setGPParameters(self, crossover_type=SINGLE_CROSSOVER,population_size=5000,generation_count=10,crossover_percentage=90):
        self.crossover_type = crossover_type
        self.population_size = population_size
        self.generation_count = generation_count
        self.crossover_percentage = crossover_percentage
        crossover_count = int(math.ceil(float(self.population_size) * float(self.crossover_percentage) / 100.0))
        reproduction_count = self.population_size - crossover_count
        if (not self.generations):
            self.generations = IterativeGeneration(self.population_size, reproduction_count, crossover_count, self.crossover_type)
        else:
            self.generations.setParameters(self.population_size, reproduction_count, crossover_count, self.crossover_type)
        self.slotModify(TRUE)
    
    def setPathName(self, fileName):
        self._fileName=fileName
        self.setTitle(str(QFileInfo(fileName).fileName()))
    
    def pathName(self):
        return self._fileName
    
    def setTitle(self, title):
        self._title=title
        self.emit(PYSIGNAL("sigDocTitleChanged"), (self._title,))

    def title(self):
        return self._title

    def newDocument(self):
        self.slotModify(FALSE)

    def open(self, fileName, format=None):
        f = open(str(fileName), 'r') # note conversion
        self._generated_generations = cPickle.load(f)
        self._viewing_generation = cPickle.load(f)
        self._viewing_individual = cPickle.load(f)
        self._robot_position = cPickle.load(f)
        
        self.crossover_type = cPickle.load(f)
        self.population_size = cPickle.load(f)
        self.generation_count = cPickle.load(f)
        self.crossover_percentage = cPickle.load(f)
        
        self.generations = cPickle.load(f)
        
        self._loaded_depth_list = cPickle.load(f)
        
        self._individual_list = cPickle.load(f)
        
        f.close()
        self.slotModify(FALSE)
        self.setPathName(fileName)
        
        self.settings_dialog = SettingsDialog()
        self.settings_dialog.restoreSettings(self.crossover_type, self.population_size, self.generation_count, self.crossover_percentage)
    
    def slotModify(self, value=None):
        if value==None:
            self._modified=not self._modified
        else:
            self._modified = value
        self.emit(PYSIGNAL("sigDocModified"),
                  (self._modified,))
    
    def modified(self):
        return self._modified
    
    def close(self):
        pass
    
    def save(self, fileName = None, format = None):
        if fileName is not None and fileName <> "":
            self.setPathName(fileName)
        else:
            if self.pathName() == None:
                raise "Could not save document: no filename."
        print "Saving to", self.pathName()
        fileName = str(self.pathName())
        f = open(fileName, 'w')
        
        cPickle.dump( self._generated_generations, f )
        cPickle.dump( self._viewing_generation, f )
        cPickle.dump( self._viewing_individual, f )
        cPickle.dump( self._robot_position, f )
        
        cPickle.dump( self.crossover_type, f )
        cPickle.dump( self.population_size, f )
        cPickle.dump( self.generation_count, f )
        cPickle.dump( self.crossover_percentage, f )
        
        cPickle.dump( self.generations, f )
        
        cPickle.dump( self._main_generation_graph._depth_list, f )
        
        cPickle.dump( self._individual_list, f )
        
        f.close()
        self.slotModify(FALSE)
